#include "SVMsubspace.h"
SVMmatrices Smatrices ; /* sparse and dense versions of subspace matrices */

int SVMsubspace /* return:
                   0 (error tolerance satisfied)
                   1 (number of iterations exceeded limit in SVM)
                   2 (out of memory)
                   3 (line search failed)
                   4 (search direction not a descent direction)
                   5 (computation of search direction fails)
                   6 (search direction not a descent direction )
                   7 (computation of mu fails)
                   8 (function value is nan in line search)*/
(
    double       *x, /* output: the solution */
    float       *Ax, /* numerical values in matrix */
    short       *Ai, /* row indices */
    int         *Ap, /* pointers to start of each column in Ax, Ai */
    short        *a, /* label vector (+- 1) */
    double        C, /* constraint is 0 <= x_i <= C */
    int           n, /* problem dimension */
    int    nfeature, /* number of features */
    short    kernel, /* 0 = linear, 1 = poly, 2 = rbf, 3 = sig */
    double      tol, /* bound on KKT error */
    SVMstat   *Stat, /* structure with statistics (can be NULL) */
    SVMparm  *sparm, /* SVM parameters, NULL = use default parameters */
    ASLparm  *aparm  /* ASL parameters, NULL = use default parameters */
)
{
    short Ax_is_1, FormS, *Coli, *NormAi, *normAi, *Si ;
    int bsub, it, j, k, l, m, num, nold, nsub, nz,
        nfree, nstore, outer_its, PrintLevel, status,
        *i, *iold, *Sp, *iWork, *ifree, *istore ;
    float *Colx, *S, *Sx, *SP ;
    double b, B, lo, hi, err, lambda0, lambda_k, nu, normg, normx, t, tic,
          SubErr, stol, serr1, serr2, ForceFac,
         *d, *g, *Sg, *SgOld, *NormAx, *normAx, *y, *yOld, *z, *xWork ;
    ASLstat ASLStat ;
    SVMcom Com ;
    SVMparm SparmStruc, *Sparm ;
    ASLparm AparmStruc, *Aparm ;

/*print_matrix (Ai, Ap, Ax, "Ax", n) ;*/

    if ( sparm == NULL )
    {
        Sparm = &SparmStruc ;
        SVMdefault (Sparm, nfeature) ;
        Sparm->gamma = 1/ (double) nfeature ;
    }
    else
    {
        Sparm = sparm ;
    }
    /* subspace error tolerance factors */
    serr1 = Sparm->serr1 ;
    serr2 = Sparm->serr2 ;
    FormS = Sparm->FormS ;
    ForceFac = Sparm->ForceFac ;

    PrintLevel = Sparm->PrintLevel ;
    if ( Sparm->PrintParms ) SVMprint_parms (Sparm) ;

    if ( aparm == NULL )
    {
        Aparm = &AparmStruc ;
        ASLdefault (Aparm) ;
    }
    else
    {
        Aparm = aparm ;
    }

    /* store variables in Com structure */
    Com.n = n ;
    Com.kernel = kernel ;
    Com.tol = tol ;
    Com.C = C ;
    Com.Ax = Ax ;
    Com.Ai = Ai ;
    Com.Ap = Ap ;
    Com.nfeature = nfeature ;
    Com.Parm = Sparm ;

    if ( Stat != NULL )
    {
        Stat->outer_its = 0 ;
        Stat->inner_its = 0 ;
        Stat->nf = 0 ;
        Stat->ng = 0 ;
        Stat->ncols = 0 ;
        Stat->nmu = 0 ;
        Stat->f = SVMZERO ;
        Stat->Subspace = SVMZERO ;
        Stat->ASL = SVMZERO ;
        Stat->BuildS = SVMZERO ;
        Stat->UpdateGrad = SVMZERO ;
    }

    /* upper bound on size of subspace */
    bsub = Sparm->nsub ;
    Com.bsub = bsub ;
    Ax_is_1 = Sparm->Ax_is_1 ;

/* ========== allocations ========== */

    /* space needed in subspace computation:
        integer: iy, iw1, iw2 (add 1 to malloc due to minSort)
        double:   y, d, z double */
    Com.xWorkSubspace = (double *) malloc (3*n*sizeof (double)) ;
    Com.iWorkSubspace = (int *) malloc ((3*n + 1)*sizeof (double)) ;

    /* space needed in gradient update computation:
        float: temp (bsub)
        integer: pget, pput, kput, kgetS, zi (bsub), OldCols, kget (bsub+1)
        double: z (bsub), LinWork (nfeature) */
    if ( kernel == 0 )
    {
        /* LinWork */
        Com.xWorkUpdate = (double *) malloc (nfeature*sizeof (double)) ;
        /* kput and kget in buildS */
        Com.iWorkUpdate = (int *) malloc ((2*bsub+1)*sizeof (int)) ;
        Com.Astore = NULL ;
    }
    else
    {
        Com.xWorkUpdate = (double *) malloc (bsub*sizeof (double)) ;
        Com.iWorkUpdate = (int *) malloc ((7*bsub+2)*sizeof (int)) ;
        Com.Astore = (float *) malloc (((long) bsub * (long) n)*sizeof (float));
    }

    /* needed to build subspace */
    Com.ikeep = (short *) malloc (n*sizeof (short)) ;
    SVMint_init (Com.ikeep, 0, n) ;

    /* space needed to solve the subspace problem and update gradient */
    g = (double *) malloc (n*sizeof (double)) ;
    d = (double *) malloc (bsub*sizeof (double)) ;
    y = (double *) malloc (bsub*sizeof (double)) ;
    yOld = (double *) malloc (bsub*sizeof (double)) ;
    z = (double *) malloc (bsub*sizeof (double)) ;
    i = (int *) malloc ((bsub+2)*sizeof (int)) ; /* flags at ends of array */
    i [0] = n ;
    i = i+1 ;                                    /* be careful */
    iold = (int *) malloc ((bsub+1)*sizeof (int)) ;/* indices prior subspace */
    ifree = (int *) malloc ((bsub+1)*sizeof (int)) ; /* current free indices */
    istore = (int *) malloc ((bsub+1)*sizeof (int)) ;/* stored columns */
    nz = Ap [n] ;
    Sx = (float *) malloc (nz*sizeof (float)) ;
    Si = (short *) malloc (nz*sizeof (short)) ;
    Sp = (int *) malloc ((bsub+1)*sizeof (int)) ;
    Sg = (double *) malloc (bsub*sizeof (double)) ;
    SgOld = NULL ;
    if ( Stat != NULL ) SgOld = (double *) malloc (bsub*sizeof (double)) ;
    normAx = NULL ;
    NormAx = NULL ;
    normAi = NULL ;
    NormAi = NULL ;

    if ( Ax_is_1 )
    {
        normAi = (short *) malloc (n*sizeof (short)) ;
        Com.normAi = normAi ;
        for (j = 0; j < n; j++) normAi [j] = (short) (Ap [j+1] - Ap [j]) ;
        NormAi = (short *) malloc (bsub*sizeof (short)) ;
    }
    else
    {
        normAx = (double *) malloc (n*sizeof (double)) ;
        Com.normAx = normAx ;
        for (j = 0; j < n; j++)
        {
            normAx [j] = SVMfloatxx (Ax+Ap [j], Ap [j+1] - Ap [j]) ;
        }
        NormAx = (double *) malloc (bsub*sizeof (double)) ;
    }
    Colx = NULL ;
    Coli = NULL ;
    /* if not a sparse dot product, then column dot products computed
       by a scatter of one column: sum Ax [k]*Colx [Ai [k]] where Colx
       stores the other column. If the number of features is small enough,
       this dot product could be more efficient than a sparse dot product */
    if ( !Sparm->SparseDot )
    {
        if ( Ax_is_1 )
        {
            Coli = (short *) malloc (nfeature*sizeof (short)) ;
            for (k = 0; k < nfeature; k++) Coli [k] = 0 ;
        }
        else
        {
            Colx = (float *) malloc (nfeature*sizeof (float)) ;
            for (k = 0; k < nfeature; k++) Colx [k] = 0. ;
        }
    }
    if ( (kernel == 0) && !FormS )
    {
        Smatrices.Sx = Sx ;
        Smatrices.Sp = Sp ;
        Smatrices.Si = Si ;
        Smatrices.temp =  Com.xWorkUpdate ;
        Smatrices.nfeature = nfeature ;
        Smatrices.Ax_is_1 = Ax_is_1 ;
        Smatrices.FormS = FormS ;
        Com.Sold = NULL ;
        S = NULL ;
    }
    else
    {
        S = (float *) malloc (((long) bsub * (long) bsub)*sizeof (float)) ;
        Smatrices.Sdense = S ;
        Smatrices.FormS = SVMTRUE ;
        FormS = SVMTRUE ;
        /* space for a copy of S matrix */
        Com.Sold = (float *) malloc (((long) bsub * (long) bsub)*sizeof(float));
    }

    /* ASL workspace */
    xWork = (double *) malloc ((6*bsub + Aparm->M + 2)*sizeof(double)) ;
    iWork = (int *) malloc ((3*bsub)*sizeof (int)) ;

    if ( (xWork == NULL) || (iWork == NULL) )
    {
        printf ("Insufficient memory for specified problem dimension %e\n",
                 (double) n) ;
        return (2) ;
    }

    /* convert problem to standard form with a_i = 1 for all i
       x_i = -y_i + C if a_i = -1
       starting guess:    x_i = 0 for all i or
                          y_i = C if a_i = -1
                          x_i = y_i = 0 otherwise
       starting gradient: g_i = +1 if a_i = -1
                          g_i = -1 otherwise */
    b = 0. ;
    for (j = 0; j < n; j++)
    {
        if ( a [j] < 0 )
        {
            x [j] = C ;
            b += C ;
            g [j] = +1 ;
        }
        else
        {
            x [j] = 0 ;
            g [j] = -1 ;
        }
    }
    Com.b = b ;

    /* initial multiplier estimate, assume all components of solution free */
    nu = -SVMadd (g, n)/((double) n) ;

    /* initializations */
    lambda_k = SVMZERO ; /* implies that first subspace chosen using LP */
    SubErr = SVMZERO ;
    it = 0 ;
    nfree = 0 ;
    nold = 0 ;
    lo = 0. ;
    hi = C ;
    outer_its = Sparm->maxouter ;

    while ( it < outer_its )
    {
        it++ ;
        if ( PrintLevel > 0 ) printf ("\nIteration: %i\n", it) ;
        /* build the subspace */
        if ( Stat != NULL ) tic = SVMtimer () ;

        status = SVMbuildSubspace (i, &nsub, &err, &B, &nu, istore,
                          &nstore, ifree, nfree, SubErr, x, g, lambda_k, &Com) ;
        if ( (status == 1) && (PrintLevel > 1) )
        {
            printf ("    repeat subspace\n") ;
        }
   
        if ( Stat != NULL )
        {
            Stat->Subspace += SVMtimer () - tic ;
            Stat->err = err ;
        }
        if ( status == 0 ) break ;
        /* The indices in the subspace are i [j], 0 <= j < nsub.
           The new iterate is given by xnew = x + d where d is
           chosen to solve the problem

               min .5(x+d)'A(x+d) - 1'(x+d)
               s.t. a'(x+d) = 0, 0 <= x+d <= C, d_j = 0 outside subspace

           In our transformed coordinates, this is equivalent to

               min .5d'Ad + g'd, g = gradient at x,
               s.t. 1'(x+d) = b, 0 <= x+d <= C, d_j = 0 outside subspace

           Above B is the sum of x_j over j outside the subspace.
           Let y_j = x_j + d_j for j in the subspace, or
               d_j = y_j - x_j. Let S denote the principal submatrix
           of A corresponding to the subspace. After the shift
           d_j = y_j - x_j, the subspace problem becomes

               min .5y'Sy + (g-Sx)'y, g = gradient at x,
               s.t. 1'y = b-B, 0 <= y <= C

           This problem is solved using ASL. The returned y yields
           the components of xnew in the subspace.

           Before the call to ASL, extract the subspace columns from
               the data matrix and form the matrix S.

           After the call to ASL, update the gradient g using d

           To begin, extract the data matrix associated with the subspace */

        /* status = 2 means this is a new subspace, otherwise reuse old */
        if ( status == 2 )
        {
            l = 0 ;
            if ( Ax_is_1 ) /* only need to extract Ai */
            {
                SVMshort_gather (NormAi, normAi, i, nsub) ;
                for (k = 0; k < nsub; k++)
                {
                    Sp [k] = l ;
                    j = i [k] ;
                    m = Ap [j] ;
                    num = Ap [j+1] - m ;
                    SVMshortCopy (Si+l, Ai+m, num) ; /* compress Ai into Si */
                    l += num ;
                }
            }
            else           /* extract both Ai and Ax */
            {
                SVMdouble_gather (NormAx, normAx, i, nsub) ;
                for (k = 0; k < nsub; k++)
                {
                    Sp [k] = l ;
                    j = i [k] ;
                    m = Ap [j] ;
                    num = Ap [j+1] - m ;;
                    SVMshortCopy (Si+l, Ai+m, num) ; /* compress Ai into Si */
                    SVMfloatCopy (Sx+l, Ax+m, num) ; /* compress Ax into Sx */
                    l += num ;
                }
            }
            Sp [nsub] = l ;

            if ( FormS )
            {
                if ( Stat != NULL ) tic = SVMtimer () ;
                SVMbuildS (S, Sx, Si, Sp, Colx, Coli, NormAx, NormAi, i, nsub,
                           iold, nold, &Com) ;
                if ( Stat != NULL ) Stat->BuildS += SVMtimer () - tic ;
            }

            /* save the subspace */
            SVMintCopy (iold, i, nsub) ;
            nold = nsub ;
        }
        /* if repeating subspace, no need to build S, however, if kernel > 0
           Sx still needs to be built since it was previously compressed
           in updateGrad */
        else if ( kernel > 0 )
        {
            l = 0 ;
            if ( Ax_is_1 ) /* only need to extract Ai */
            {
                SVMshort_gather (NormAi, normAi, i, nsub) ;
                for (k = 0; k < nsub; k++)
                {
                    Sp [k] = l ;
                    j = i [k] ;
                    m = Ap [j] ;
                    num = Ap [j+1] - m ;
                    SVMshortCopy (Si+l, Ai+m, num) ; /* compress Ai into Si */
                    l += num ;
                }
            }
            else           /* extract both Ai and Ax */
            {
                SVMdouble_gather (NormAx, normAx, i, nsub) ;
                for (k = 0; k < nsub; k++)
                {
                    Sp [k] = l ;
                    j = i [k] ;
                    m = Ap [j] ;
                    num = Ap [j+1] - m ;;
                    SVMshortCopy (Si+l, Ai+m, num) ; /* compress Ai into Si */
                    SVMfloatCopy (Sx+l, Ax+m, num) ; /* compress Ax into Sx */
                    l += num ;
                }
            }
            Sp [nsub] = l ;
        }
        else if ( status == 1 ) it-- ;

        /* gather starting guess y for subspace optimization */
        SVMdouble_gather (y, x, i, nsub) ;

        /* save copy of y */
        SVMdoubleCopy (yOld, y, nsub) ;

        /* gather gradient at y for subspace objective function */
        SVMdouble_gather (Sg, g, i, nsub) ;

        /* save copy of Sg if cost is to be evaluated */
        if ( Stat != NULL ) SVMdoubleCopy (SgOld, Sg, nsub) ;

        if ( it == 1 ) /* initial estimate for lambda_k */
        {
            normg = SVMmax (g, n) ;
            normx = SVMmax (x, n) ;
            /* lower bound for lambda_k from ASLparm */
            lambda0 = Aparm->lambda0 ;
            if ( normx != SVMZERO)  lambda_k = SVMMAX (normg/normx, lambda0) ;
            else                    lambda_k = SVMMAX (normg, lambda0) ;
        }

        /* solve the subspace problem */
        if ( Stat != NULL ) tic = SVMtimer () ;

        stol = SVMMAX(serr1*tol, serr2*err) ;
        while ( 1 )
        {
            status = ASL (y, &lambda_k, &lo, &hi, NULL, b-B, Sg, nsub, stol,
                     xWork, iWork, &ASLStat, Aparm, SVMSmult, NULL, NULL, NULL);

            /* If the algorithm stops in the first iteration, reduce error
               tolerance stol and continue the iteration with a smaller
               error tolerance. */
            if ( (ASLStat.it == 1) && (status == 0) )
            {
                stol = ForceFac*ASLStat.err ;
                continue ;
            }
            else break ;
        }

        /* scatter solution back into x */
        SVMdouble_scatter (x, y, i, nsub) ;

        if ( (status != 0) && (status != 5) && (Stat == NULL) ) break ;
        if ( (status == 5) && (ASLStat.it == 1) && (Stat == NULL) ) break ;

        /* compute d = y - yOld */
        SVMdense_diff (d, y, yOld, nsub) ;

        /* compute z = Sd */
        if ( FormS )
        {
            SP = S ;
            for (j = 0; j < nsub; j++)
            {
                z [j] = SVMfloat_double_dot(SP, d, nsub) ;
                SP = SP+nsub ;
            }
        }
        else SVMsparseProduct (z, d, nsub) ;

        if ( Stat != NULL )
        {
            Stat->ASL += SVMtimer () - tic ;
            Stat->outer_its = it ;
            Stat->inner_its += ASLStat.it ;
            Stat->nf += ASLStat.nf ;
            Stat->ng += ASLStat.ng ;
            Stat->ncols += nsub - nstore ;
            Stat->nmu += ASLStat.nmu ;
            /* change in cost is (.5z + g)'d, z = Sd */
            t = 0. ;
            for (j = 0; j < nsub; j++) t += d [j]*(.5*z [j] + SgOld [j]) ;
            Stat->f += t ;
            if ( PrintLevel > 0 )
                printf ("    cost:                   %e\n", Stat->f) ;
            if ( (status != 0) && (status != 5) ) break ;
            if ( (status == 5) && (ASLStat.it == 1) ) break ;
        }

        /* extract subspace error */
        SubErr = ASLStat.err ;
        if ( PrintLevel > 1 )
            printf ("    subspace error:         %e\n", SubErr) ;

        /* add z = Sd to g */
        for (j = 0; j < nsub; j++) g [i [j]] += z [j] ;

        /* update gradient gnew = g + A*d */
        if ( Stat != NULL ) tic = SVMtimer () ;
        SVMupdateGrad (g, d, y, i, nsub, istore, nstore, ifree, &nfree,
                       S, Sx, Si, Sp, Colx, Coli, NormAx, NormAi, &Com) ;
        if ( Stat != NULL ) Stat->UpdateGrad += SVMtimer () - tic ;
    }

    /* reverse the transformation on x */
    for (j = 0; j < n; j++) if ( a [j] < 0 ) x [j] = C - x [j] ;

    if ( it == Sparm->maxouter ) status = 1 ;
    if ( Stat != NULL ) Stat->err = err ;

    /* free memory */
    if ( normAx != NULL ) free (normAx) ;
    if ( NormAx != NULL ) free (NormAx) ;
    if ( normAi != NULL ) free (normAi) ;
    if ( NormAi != NULL ) free (NormAi) ;
    free (Com.xWorkSubspace) ;
    free (Com.iWorkSubspace) ;
    free (Com.xWorkUpdate) ;
    if ( Com.iWorkUpdate != NULL ) free (Com.iWorkUpdate) ;
    if ( Com.Astore != NULL ) free (Com.Astore) ;
    free (Com.ikeep) ;
    free (g) ;
    free (d) ;
    free (y) ;
    free (yOld) ;
    free (z) ;
    free (i-1) ;
    free (iold) ;
    free (ifree) ;
    free (istore) ;
    if ( S != NULL ) free (S) ;
    if ( Com.Sold != NULL ) free (Com.Sold) ;
    free (Sx) ;
    free (Si) ;
    free (Sp) ;
    free (Sg) ;
    if ( SgOld != NULL ) free (SgOld) ;
    if ( Colx != NULL ) free (Colx) ;
    if ( Coli != NULL ) free (Coli) ;
    free (xWork) ;
    free (iWork) ;
    return (status) ;
}

/* Compute the product p = S'Sd where S is a sparse matrix */

PRIVATE void SVMsparseProduct
(
    double *p,
    double *d,
    int     n
)
{
    int j, l, nfeature, *Sp ;
    short Ax_is_1, *Si ;
    double *temp ;
    float *Sx ;

    Sx = Smatrices.Sx ;
    Sp = Smatrices.Sp ;
    Si = Smatrices.Si ;
    nfeature = Smatrices.nfeature ;
    temp = Smatrices.temp ;
    Ax_is_1 = Smatrices.Ax_is_1 ;

    SVMscalar_init (temp, 0., nfeature) ;
    if ( Ax_is_1 )
    {
        for (j = 0; j < n; j++)
        {
            l = Sp [j] ;
            SVMsparse_add (temp, Si+l, d [j], Sp [j+1] - l) ;
        }
        for (j = 0; j < n; j++)
        {
            l = Sp [j] ;
            p [j] = SVMbinary_double_dot (temp, Si+l, Sp [j+1] - l) ;
        }
    }
    else
    {
        for (j = 0; j < n; j++)
        {
            l = Sp [j] ;
            SVMsparse_daxpy (temp, Sx+l, Si+l, d [j], Sp [j+1] - l) ;
        }
        for (j = 0; j < n; j++)
        {
            l = Sp [j] ;
            p [j] = SVMdouble_dense_dot (temp, Sx+l, Si+l, Sp [j+1] - l) ;
        }
    }
}

PRIVATE void SVMSmult
(
    ASLobjective *data
)
{
    int j, k, l, n, nfeature, nz, *i, *Sp ;
    short Ax_is_1, *Si ;
    double *d, *p, *temp ;
    float *S, *Sx ;

    nz = data->nz ;
    i = data->i ;
    d = data->d ;
    p = data->p ;
    n = data->n ;

    if ( Smatrices.FormS )
    {
        S = Smatrices.Sdense ;
        SVMscalar_init (p, 0., n) ;
        for (k = 0; k < nz; k++) SVMdense_daxpy (p, S+(i [k]*n), d [k], n) ;
    }
    else
    {
        Sx = Smatrices.Sx ;
        Sp = Smatrices.Sp ;
        Si = Smatrices.Si ;
        nfeature = Smatrices.nfeature ;
        temp = Smatrices.temp ;
        Ax_is_1 = Smatrices.Ax_is_1 ;

        SVMscalar_init (temp, 0., nfeature) ;
        if ( Ax_is_1 )
        {
            for (k = 0; k < nz; k++)
            {
                j = i [k] ;
                l = Sp [j] ;
                SVMsparse_add (temp, Si+l, d [k], Sp [j+1] - l) ;
            }
            for (j = 0; j < n; j++)
            {
                l = Sp [j] ;
                p [j] = SVMbinary_double_dot (temp, Si+l, Sp [j+1] - l) ;
            }
        }
        else
        {
            for (k = 0; k < nz; k++)
            {
                j = i [k] ;
                l = Sp [j] ;
                SVMsparse_daxpy (temp, Sx+l, Si+l, d [k], Sp [j+1] - l) ;
            }
            for (j = 0; j < n; j++)
            {
                l = Sp [j] ;
                p [j] = SVMdouble_dense_dot (temp, Sx+l, Si+l, Sp [j+1] - l) ;
            }
        }
    }
}

/*  The first step in the subspace calculation is to estimate the
    multiplier nu associated with the linear constraint in the bound and
    linearly constrained problem

    (BL) min f(x)  subject to  0 <= x <= C,  a'x = b.

    We use two different approaches depending on the size of the BB
    estimate lambda_k for the Hessian of (BL). On approach is based on
    the quadratic program

    (QP) min .5*lambda_k ||d||^2 + g'd subject to  lo <= x + d <= hi, a'd = 0,

    where g = grad f(x).  However, if lambda_k is tiny, then there
    are rounding error problems when we try to solve this quadratic program.
    When lambda_k is sufficiently small, we drop the quadratic term and
    consider the linear program

    (LP) min g'd subject to 0 <= x + d <= C, a'd = 0.

    For the SVM problem, we use (LP) when

         max_i |g_i| >= lpfactor*C*lambda_k

    where the constraints for SVM are 0 <= x <= C.

    To solve (QP) or (LP), we make the change of variables
    z = x + d. This leads to the equivalent problems

    (QP') min ||z-y|| subject to lo <= z <= hi, a'z = b

    where y = x - g/lambda_k and b = a'x, and

    (LP') min g'z subject to lo <= z <= hi, a'z = b

    At a solution of (BL), the multiplier nu associated with the linear
    constraint is the same as the multiplier associated with linear
    constraints for (QP), (LP), and (LP'). However, for (QP') the multiplier
    mu for the linear constraint satisfies nu = lambda_k*mu.

    From the multiplier estimate, we obtain an estimate for the error
    in a approximate solution to (LB) by computing the smallest perturbation
    that allows us to satisfy the KKT conditions

         (g_i + nu) =  0 if 0 < x_i < C,
         (g_i + nu) <= 0 if x_i = C,
         (g_i + nu) >= 0 if x_i = 0.

    We have the following ways to perturb the KKT conditions:

         (g_i + p_i + nu) =  0 if 0 < x_i < C
         (g_i + p_i + nu) <= 0 if x_i + q_i = C
         (g_i + p_i + nu) >= 0 if x_i + q_i = 0

    If the error estimate satisfies the stopping conditions, then we stop.
    If the error in the solution to the previous subspace problem is too
    big relative to the current error, then we return to the previous
    subspace problem and further refine the error.

    The subspace is chosen in the following way. We consider the
    individual terms in the Lagrangian for either (QP) or (LP):

    t_i = .5*lambda_k d_i^2 + (g_i + nu)d_i or t_i = (g_i + nu)d_i

    Let J and K be indices corresponding to d_i > 0 and d_i < 0 respectively.
    Arrange the indices in J and K so that the t_i for i \in J or
    t_i for i \in K are in increasing order. Start with the first
    element in J or K and put it in the subspace. Then go to the
    opposite list and add indices until the d_i sum changes sign.
    Repeat this process until obtaining the desired number of indices <=
    alpha*bsub, where alpha denotes the fraction of the subspace indices
    chosen in this first phase (alpha = 2/3 default). In the second
    phase, add indices from the previous subspace until obtaining bsub
    indices if possible.  Only add indices corresponding to free variables,
    with preference to indices for which the absolute gradient is largest. */

PRIVATE int SVMbuildSubspace /* return:
                   0 (error tolerance satisfied)
                   1 (refine accuracy for prior subspace)
                   2 (subspace has been evaluated) */
(
    int          *i, /* indices of subspace (returned) */
    int       *nsub, /* size of subspace (returned) */
    double     *Err, /* error estimate (returned) */
    double       *B, /* 1'x outside subsubspace (returned) */
    double      *nu, /* multiplier estimate (input and returned) */
    int     *istore, /* stored column in current subspace */
    int     *nstore, /* number of stored columns in current subspace */
    int      *ifree, /* free indices from previous subspace */
    int       nfree, /* number of free indices from previous subspace */
    double   SubErr, /* error for previous subspace */
    double       *x, /* current estimate of solution to problem */
    double       *g, /* gradient at current point */
    double lambda_k, /* current BB parameter */
    SVMcom     *Com  /* SVM Com structure */
)
{
    short kernel, qp, take1, *ikeep ;
    int I1, I2, j, k, kj, n, nn, np, npre, MaxKept, Nsub, Nsub1, Nsub2,
        Nstore, PrintLevel, top1, top2, *iy, *i1, *iw1, *i2, *iw2, *p1, *p2 ;
    double b, C, d_pert, dsum, err, ltol, lpfactor, mu, Nu, PreAmp, r, s, t,
           u, xj, Y, tol, *y, *d, *d1, *d2, *z, *z1, *z2 ;
    SVMparm *Parm ;

    /* ======== extract variables ======== */
    n = Com->n ;
    C = Com->C ;
    y = Com->xWorkSubspace ;
    d = y+n ;
    z = d+n ;
    iy = Com->iWorkSubspace ;
    iw1 = iy+n ;
    iw2 = iw1+n ;
    kernel = Com->kernel ;
    tol = Com->tol ;
    b = Com->b ;
    ikeep = Com->ikeep ; /* array initialized to be zero */

    Parm = Com->Parm ;
    PrintLevel = Parm->PrintLevel ;
    d_pert = Parm->d_pert ;
    ltol = tol*Parm->l_pert ;
    lpfactor = Parm->lpfactor ;
    Nsub = SVMMIN (Parm->nsub, n) ;
    MaxKept = Nsub*Parm->MaxKept ;
    PreAmp = Parm->PreAmp ;
    /* ======== end of extract ======== */

    /* Decide whether the multiplier and z = x + d will be computed
       by solving QP or LP */
    t = SVMmax (g, n) ;
    qp = 0 ;
    if ( t >= lpfactor*C*lambda_k ) /* solve LP */
    {
        Nu = SVMlp (y, g, C, b, n, *nu, iw1, iw2, z) ;
    }
    else                        /* solve QP */
    {
        SVMstep (y, x, g, -SVMONE/lambda_k, n) ;
        {
            mu = *nu/lambda_k ;
            mu = SVMnapsack (y, mu, b, SVMZERO, C, n, z, iw1, iw2) ;
            Nu = lambda_k*mu ;
            qp = 1 ;
        }
    }
    *nu = Nu ;
    /* y now stores z, compute the following:
       1. error estimate based on the smallest change
          in g_j or xj for which optimality conditions hold
       2. d = y - x
       3. the subspace is based on the smallest elements of
          .5 lambda_k d_j^2 + (g_j + nu a_j)d_j or
                              (g_j + nu a_j)d_j */

    /* flag prior free indices */
    for (k = 0; k < nfree; k++) ikeep [ifree [k]] = -1 ;
    err = SVMZERO ;
    s = .5*lambda_k ;
    nn = 0 ;
    np = n ;
    for (j = 0; j < n; j++)
    {
        xj = x [j] ;
        t = y [j] - xj ;
        r = g [j] + Nu ;
        if ( fabs (t) >= d_pert ) /* include in d */
        {
            if ( qp ) u = t*(s*t + r) ;
            else      u = t*r ;
            if ( t > SVMZERO )
            {
                np-- ;
                if ( ikeep [j] ) z [np] = PreAmp*u ;
                else             z [np] = u ;
                iy [np] = j ;
                d [np] = t ;
            }
            else
            {
                if ( ikeep [j] ) z [nn] = PreAmp*u ;
                else             z [nn] = u ;
                iy [nn] = j ;
                d [nn] = t ;
                nn++ ;
            }
        }
        if ( r > SVMZERO ) /* min { r, x [j] } */
        {
            if ( r > xj ) err = SVMMAX (err, xj) ;
            else          err = SVMMAX (err, r) ;
        }
        else               /* min { -r, C - x [j] } */
        {
            Y = C - xj ;
            if (-r > Y ) err = SVMMAX (err, Y) ;
            else         err = SVMMAX (err,-r) ;
        }
    }
    *Err = err ;
     if ( PrintLevel > 0 ) printf ("    error:                  %e\n", err) ;

    /* check if termination condition is satisfied */
    if ( err <= tol ) return (0) ;

    /* check whether the prior subspace should be repeated */
    if ( SubErr > err )
    {
        /* If kernel > 0, then the columns associated with the
           previous free indices have been evaluated. If kernel = 0,
           then the kernel is linear and we do not save the columns */
        if ( kernel > 0 )
        {
            SVMintCopy (istore, ifree, nfree) ;
            *nstore = nfree ;
        }
        else *nstore = 0 ;
        return (1) ;
    }

    if ( Nsub >= nn + (n - np) ) /* put all the indices in the subspace */
    {
        npre = nfree ;
        for (k = 0; k < nn; k++)
        {
            j = iy [k] ;
            i [k] = j ;
            ikeep [j] = 1 ;
        }
        Nsub = nn ;
        for (k = np; k < n; k++)
        {
            j = iy [k] ;
            i [Nsub] = j ;
            Nsub++ ;
            ikeep [j] = 1 ;
        }
    }

    else                        /* need to choose indices for subspace */
    {
        /* find Nsub smallest elements in the bottom part of z */
        Nsub1 = SVMMIN(Nsub, nn) ;
        SVMpartialMinSort (iw2, z, Nsub1, nn) ;
        SVMminHeapSort (iw1, iw2, z, Nsub1) ;

        /* find Nsub smallest elements in the top part of z */
        Nsub2 = SVMMIN(Nsub, n-np) ;
        SVMpartialMinSort (iw2+np, z+np, Nsub2, n-np) ;
        SVMminHeapSort (iw1+np, iw2+np, z+np, Nsub2) ;

        z1 = z ;     /* cost improvement when d < 0 */
        i1 = iw1 ;   /* indices giving increasing order in z1 */
        p1 = iy ;    /* original indices */
        d1 = d ;     /* direction y - x */

        z2 = z+np ;  /* cost improvement when d > 0 */
        i2 = iw1+np ;/* indices giving increasing order in z2 */
        p2 = iy+np ; /* original indices */
        d2 = d+np ;  /* direction y - x */

        npre = 0 ;   /* # of previous free subspace indices in new subspace */
        k = 0 ;
        top1 = 0 ;
        top2 = 0 ;
        I1 = i1 [0] ;
        I2 = i2 [0] ;
        dsum = SVMZERO ;

        take1 = SVMTRUE ;
        if ( z1 [I1] > z2 [I2] ) take1 = SVMFALSE ;
        while ( k < Nsub )
        {
            if ( take1 )
            {
                I1 = i1 [top1] ;
                top1++ ;
                j = p1 [I1] ;
                kj = ikeep [j] ;
                if ( !kj || (kj && (npre < MaxKept)) )
                {
                    if ( kj ) npre++ ;
                    dsum += d1 [I1] ;
                    i [k] = j ;
                    ikeep [j] = 1 ;
                    k++ ;
                }
            }
            else
            {
                I2 = i2 [top2] ;
                top2++ ;
                j = p2 [I2] ;
                kj = ikeep [j] ;
                if ( !kj || (kj && (npre < MaxKept)) )
                {
                    if ( kj ) npre++ ;
                    dsum += d2 [I2] ;
                    i [k] = j ;
                    ikeep [j] = 1 ;
                    k++ ;
                }
            }
            if ( (top1 == Nsub1) || (top2 == Nsub2) ) break ;
            if      ( dsum > SVMZERO ) take1 = SVMTRUE ;
            else if ( dsum < SVMZERO ) take1 = SVMFALSE ;
            else
            {
                if ( z1 [I1] < z2 [I2] ) take1 = SVMTRUE ;
                else                     take1 = SVMFALSE ;
            }
        }
        if ( k < Nsub )
        {
            if ( top1 == Nsub1 )
            {
                while ( (k < Nsub) && (top2 < Nsub2) )
                {
                    j = p2 [i2 [top2]] ;
                    top2++ ;
                    kj = ikeep [j] ;
                    if ( !kj || (kj && (npre < MaxKept)) )
                    {
                        if ( kj ) npre++ ;
                        i [k] = j ;
                        ikeep [j] = 1 ;
                        k++ ;
                    }
                }
            }
            else
            {
                while ( (k < Nsub) && (top1 < Nsub1) )
                {
                    j = p1 [i1 [top1]] ;
                    top1++ ;
                    kj = ikeep [j] ;
                    if ( !kj || (kj && (npre < MaxKept)) )
                    {
                        if ( kj ) npre++ ;
                        i [k] = j ;
                        ikeep [j] = 1 ;
                        k++ ;
                    }
                }
            }
        }
        Nsub = k ;
    }

    if (PrintLevel > 1)
    {
        printf ("    prior subspace indices: %i\n", npre) ;
        printf ("    total subspace size:    %i\n", Nsub) ;
    }

    /* At this point, the subspace is complete.
       Nsub = nsub = dimension of current subspace */
    *nsub = Nsub ;

    /* Remove 1s from ikeep and compute a'x outside subspace.
       For nonlinear kernels, we also determine which of the
       indices in the subspace correspond to columns that were
       saved in the previous iteration.  These indices are placed
       in the array istore. When we update the gradient, we do not
       need to evaluate these columns since there are in Astore */
    b = SVMZERO ;
    if ( kernel > 0 ) /* only save columns for nonlinear kernels */
    {
        k = 0 ;
        Nstore = 0 ;
        ifree [nfree] = n ;
        for (j = 0; j < n; j++)
        {
            if ( ikeep [j] > 0 )      /* this index is in current subspace */
            {
                while ( ifree [k] < j ) k++ ;
                if ( ifree [k] == j ) /* this index was in prior subspace */
                {
                    istore [Nstore] = j ;
                    Nstore++ ;
                }
            }
            /* b stores a'x outside subspace */
            else b += x [j] ;
            ikeep [j] = 0 ;
        }
        *nstore = Nstore ;
    }
    else
    {
        for (j = 0; j < n; j++)
        {
            /* b stores a'x outside subspace */
            if ( ikeep [j] <= 0 ) b += x [j] ;
            ikeep [j] = 0 ;
        }
        *nstore = 0 ;
    }
    *B = b ;
    if ( PrintLevel > 1 ) printf ("    number cache cols:      %i\n", *nstore);

    /* sort subspace indices in increasing order */
    SVMminSort (i, iw1, iw2, Nsub) ;

    return (2) ;
}

PRIVATE void SVMbuildS
(
    float       *S, /* svm matrix corresponding to subspace */
    float      *Sx, /* numerical data */
    short      *Si, /* features associated with data */
    int        *Sp, /* pointer array */
    float    *Colx, /* storage for float column of data */
    short    *Coli, /* storage for 1s column of data */
    double *NormAx, /* norm of each column, columns are float */
    short  *NormAi, /* norm of each column, columns are 1s */
    int         *I, /* subspace indices */
    int       nsub, /* size of I */
    int      *Iold, /* indices for prior subspace */
    int       nold, /* size of Iold */
    SVMcom    *Com
)
{
    short Ax_is_1, degree, kernel, SparseDot, *Sii ;
    int i, j, jp, k, knum, kp, l, n, nzx, nzy, si, O, PrintLevel,
       *kget, *kput, *jput ;
    float *SP, *Sold, *SPold, *Sxi ;
    double gamma, t, sx ;
    SVMparm *Parm ;

    /* ======== extract variables ======== */
    n = Com->n ;
    Sold = Com->Sold ;
    kernel = Com->kernel ;
    kget = Com->iWorkUpdate ;
    kput = kget+(Com->bsub) ;

    Parm = Com->Parm ;
    PrintLevel = Parm->PrintLevel ;
    degree = Parm->degree ;
    gamma = Parm->gamma ;
    Ax_is_1 = Parm->Ax_is_1 ;
    SparseDot = Parm->SparseDot ;
    /* ======== end of extract ======== */

    /* if nold > 0, find the subspaces indices of Iold which are also in I */
    l = 0 ;
    if ( nold > 0 )
    {
        Iold [nold] = n ;
        k = 0 ;
        for (j = 0; j < nsub; j++)
        {
            i = I [j] ;
            while ( Iold [k] < i ) k++ ;
            if ( Iold [k] == i )
            {
                kget [l] = k ;
                k++ ;
                kput [l] = j ;
                l++ ;
            }
        }
        /* l = number of indices from current subspace that are contained
               in the prior subspace
           kget [l] = column location in old S
           kput [l] = column location in new S
           copy from  S and compress into Sold: */
        SPold = Sold ;
        for (k = 0; k < l; k++)
        {
            SP = S+nold*kget [k] ;
            /* only copy element on right side of main diagonal */
            SVMfloat_gather (SPold, SP, kget+k, l-k) ;
            SPold = SPold+(l-k) ;
        }
    }
    kput [l] = n ;
    knum = l ;
    if ( PrintLevel > 1 ) printf ("    dimension of reused S:  %i\n", knum) ;

    SPold = Sold ;
    jp = n ;
    kp = *kput ;
    k = 0 ;
    if ( Ax_is_1 )
    {
        if ( SparseDot )
        {
            if ( kernel == 0 )
            {
                for (i = 0; i < nsub; i++)
                {
                    if ( kp == i ) /* copy elements from Sold */
                    {
                        SVMfloat_scatter (S+k, SPold, kput, knum) ;
                        SPold = SPold+knum ;
                        knum-- ;
                        kput = kput+1 ;
                        kp = *kput ;
                        jp = kp ;
                        jput = kput ;
                        k += i + 1 ;
                    }
                    else
                    {
                        k += i ;
                        S [k] = NormAi [i] ;
                        k++ ;
                    }
                    nzx = Sp [i+1] - Sp [i] ;
                    Sii = Si+Sp [i] ;
                    for (j = i+1; j < nsub; j++)
                    {
                        if ( j == jp ) /* this element stored in scatter above*/
                        {
                            jput = jput+1 ;
                            jp = *jput ;
                            k++ ;
                            continue ;
                        }
                        nzy = Sp [j+1] - Sp [j] ;
                        S [k] = SVMbinary_sparse_dot (Sii, nzx, Si+Sp[j], nzy) ;
                        k++ ;
                    }
                }
            }
            else if ( kernel == 1 )
            {
                for (i = 0; i < nsub; i++)
                {
                    if ( kp == i ) /* copy elements from Sold */
                    {
                        SVMfloat_scatter (S+k, SPold, kput, knum) ;
                        SPold = SPold+knum ;
                        knum-- ;
                        kput = kput+1 ;
                        kp = *kput ;
                        jp = kp ;
                        jput = kput ;
                        k += i + 1 ;
                    }
                    else
                    {
                        k += i ;
                        t = gamma*NormAi [i] ;
                        if ( degree == 3 ) t = t*t*t ;
                        else if ( degree == 2 )   t = t*t ;
                        else if ( degree == 4 ) { t = t*t; t = t*t ; }
                        else                      t = pow(t, degree) ;
                        S [k] = t ;
                        k++ ;
                    }
                    nzx = Sp [i+1] - Sp [i] ;
                    Sii = Si+Sp [i] ;
                    for (j = i+1; j < nsub; j++)
                    {
                        if ( j == jp ) /* this element stored in scatter above*/
                        {
                            jput = jput+1 ;
                            jp = *jput ;
                            k++ ;
                            continue ;
                        }
                        nzy = Sp [j+1] - Sp [j] ;
                        t = gamma*SVMbinary_sparse_dot(Sii, nzx, Si+Sp[j], nzy);
                        if ( degree == 3 ) t = t*t*t ;
                        else if ( degree == 2 )   t = t*t ;
                        else if ( degree == 4 ) { t = t*t; t = t*t ; }
                        else                      t = pow(t, degree) ;
                        S [k] = t ;
                        k++ ;
                    }
                }
            }
            else if ( kernel == 2 )
            {
                for (i = 0; i < nsub; i++)
                {
                    si = NormAi [i] ;
                    if ( kp == i ) /* copy elements from Sold */
                    {
                        SVMfloat_scatter (S+k, SPold, kput, knum) ;
                        SPold = SPold+knum ;
                        knum-- ;
                        kput = kput+1 ;
                        kp = *kput ;
                        jp = kp ;
                        jput = kput ;
                        k += i + 1 ;
                    }
                    else
                    {
                        k += i ;
                        S [k] = 1 ;
                        k++ ;
                    }
                    nzx = Sp [i+1] - Sp [i] ;
                    Sii = Si+Sp [i] ;
                    for (j = i+1; j < nsub; j++)
                    {
                        if ( j == jp ) /* this element stored in scatter above*/
                        {
                            jput = jput+1 ;
                            jp = *jput ;
                            k++ ;
                            continue ;
                        }
                        nzy = Sp [j+1] - Sp [j] ;
                        t = SVMbinary_sparse_dot (Sii, nzx, Si+Sp[j], nzy) ;
                        S [k] = exp (gamma*(2.*t - si - NormAi [j])) ;
                        k++ ;
                    }
                }
            }
            else /* kernel == 3 */
            {
                for (i = 0; i < nsub; i++)
                {
                    if ( kp == i ) /* copy elements from Sold */
                    {
                        SVMfloat_scatter (S+k, SPold, kput, knum) ;
                        SPold = SPold+knum ;
                        knum-- ;
                        kput = kput+1 ;
                        kp = *kput ;
                        jp = kp ;
                        jput = kput ;
                        k += i + 1 ;
                    }
                    else
                    {
                        k += i ;
                        S [k] = tanh (gamma*(double) NormAi [i]) ;
                        k++ ;
                    }
                    nzx = Sp [i+1] - Sp [i] ;
                    Sii = Si+Sp [i] ;
                    for (j = i+1; j < nsub; j++)
                    {
                        if ( j == jp ) /* this element stored in scatter above*/
                        {
                            jput = jput+1 ;
                            jp = *jput ;
                            k++ ;
                            continue ;
                        }
                        nzy = Sp [j+1] - Sp [j] ;
                        t = SVMbinary_sparse_dot (Sii, nzx, Si+Sp[j], nzy) ;
                        S [k] = tanh (gamma*t) ;
                        k++ ;
                    }
                }
            }
        }
        else /* not SparseDot */
        {
            if ( kernel == 0 )
            {
                for (i = 0; i < nsub; i++)
                {
                    if ( kp == i ) /* copy elements from Sold */
                    {
                        SVMfloat_scatter (S+k, SPold, kput, knum) ;
                        SPold = SPold+knum ;
                        knum-- ;
                        kput = kput+1 ;
                        kp = *kput ;
                        jp = kp ;
                        jput = kput ;
                        k += i + 1 ;
                    }
                    else
                    {
                        k += i ;
                        S [k] = NormAi [i] ;
                        k++ ;
                    }
                    l = Sp [i+1] ;
                    for (j = Sp [i]; j < l; j++) Coli [Si [j]] = 1 ;
                    for (j = i+1; j < nsub; j++)
                    {
                        if ( j == jp ) /* this element stored in scatter above*/
                        {
                            jput = jput+1 ;
                            jp = *jput ;
                            k++ ;
                            continue ;
                        }
                        nzy = Sp [j+1] - Sp [j] ;
                        S [k] = SVMbinary_short_dot (Coli, Si+Sp[j], nzy) ;
                        k++ ;
                    }
                    for (j = Sp [i]; j < l; j++) Coli [Si [j]] = 0 ;
                }
            }
            else if ( kernel == 1 )
            {
                for (i = 0; i < nsub; i++)
                {
                    if ( kp == i ) /* copy elements from Sold */
                    {
                        SVMfloat_scatter (S+k, SPold, kput, knum) ;
                        SPold = SPold+knum ;
                        knum-- ;
                        kput = kput+1 ;
                        kp = *kput ;
                        jp = kp ;
                        jput = kput ;
                        k += i + 1 ;
                    }
                    else
                    {
                        k += i ;
                        t = gamma*NormAi [i] ;
                        if ( degree == 3 ) t = t*t*t ;
                        else if ( degree == 2 )   t = t*t ;
                        else if ( degree == 4 ) { t = t*t; t = t*t ; }
                        else                      t = pow(t, degree) ;
                        S [k] = t ;
                        k++ ;
                    }
                    l = Sp [i+1] ;
                    for (j = Sp [i]; j < l; j++) Coli [Si [j]] = 1 ;
                    for (j = i+1; j < nsub; j++)
                    {
                        if ( j == jp ) /* this element stored in scatter above*/
                        {
                            jput = jput+1 ;
                            jp = *jput ;
                            k++ ;
                            continue ;
                        }
                        nzy = Sp [j+1] - Sp [j] ;
                        t = gamma*SVMbinary_short_dot (Coli, Si+Sp[j], nzy) ;
                        if ( degree == 3 ) t = t*t*t ;
                        else if ( degree == 2 )   t = t*t ;
                        else if ( degree == 4 ) { t = t*t; t = t*t ; }
                        else                      t = pow(t, degree) ;
                        S [k] = t ;
                        k++ ;
                    }
                    for (j = Sp [i]; j < l; j++) Coli [Si [j]] = 0 ;
                }
            }
            else if ( kernel == 2 )
            {
                for (i = 0; i < nsub; i++)
                {
                    si = NormAi [i] ;
                    if ( kp == i ) /* copy elements from Sold */
                    {
                        SVMfloat_scatter (S+k, SPold, kput, knum) ;
                        SPold = SPold+knum ;
                        knum-- ;
                        kput = kput+1 ;
                        kp = *kput ;
                        jp = kp ;
                        jput = kput ;
                        k += i + 1 ;
                    }
                    else
                    {
                        k += i ;
                        S [k] = 1 ;
                        k++ ;
                    }
                    l = Sp [i+1] ;
                    for (j = Sp [i]; j < l; j++) Coli [Si [j]] = 1 ;
                    for (j = i+1; j < nsub; j++)
                    {
                        if ( j == jp ) /* this element stored in scatter above*/
                        {
                            jput = jput+1 ;
                            jp = *jput ;
                            k++ ;
                            continue ;
                        }
                        nzy = Sp [j+1] - Sp [j] ;
                        t = SVMbinary_short_dot (Coli, Si+Sp[j], nzy) ;
                        S [k] = exp (gamma*(2.*t - si - NormAi [j])) ;
                        k++ ;
                    }
                    for (j = Sp [i]; j < l; j++) Coli [Si [j]] = 0 ;
                }
            }
            else /* kernel == 3 */
            {
                for (i = 0; i < nsub; i++)
                {
                    if ( kp == i ) /* copy elements from Sold */
                    {
                        SVMfloat_scatter (S+k, SPold, kput, knum) ;
                        SPold = SPold+knum ;
                        knum-- ;
                        kput = kput+1 ;
                        kp = *kput ;
                        jp = kp ;
                        jput = kput ;
                        k += i + 1 ;
                    }
                    else
                    {
                        k += i ;
                        S [k] = tanh (gamma*(double) NormAi [i]) ;
                        k++ ;
                    }
                    l = Sp [i+1] ;
                    for (j = Sp [i]; j < l; j++) Coli [Si [j]] = 1 ;
                    for (j = i+1; j < nsub; j++)
                    {
                        if ( j == jp ) /* this element stored in scatter above*/
                        {
                            jput = jput+1 ;
                            jp = *jput ;
                            k++ ;
                            continue ;
                        }
                        nzy = Sp [j+1] - Sp [j] ;
                        t = SVMbinary_short_dot (Coli, Si+Sp[j], nzy) ;
                        S [k] = tanh (gamma*t) ;
                        k++ ;
                    }
                    for (j = Sp [i]; j < l; j++) Coli [Si [j]] = 0 ;
                }
            }
        }
    }
    else /* Ax != 1 */
    {
        if ( SparseDot )
        {
            if ( kernel == 0 )
            {
                for (i = 0; i < nsub; i++)
                {
                    if ( kp == i ) /* copy elements from Sold */
                    {
                        SVMfloat_scatter (S+k, SPold, kput, knum) ;
                        SPold = SPold+knum ;
                        knum-- ;
                        kput = kput+1 ;
                        kp = *kput ;
                        jp = kp ;
                        jput = kput ;
                        k += i + 1 ;
                    }
                    else
                    {
                        k += i ;
                        S [k] = NormAx [i] ;
                        k++ ;
                    }
                    nzx = Sp [i+1] - Sp [i] ;
                    Sii = Si+Sp [i] ;
                    Sxi = Sx+Sp [i] ;
                    for (j = i+1; j < nsub; j++)
                    {
                        if ( j == jp ) /* this element stored in scatter above*/
                        {
                            jput = jput+1 ;
                            jp = *jput ;
                            k++ ;
                            continue ;
                        }
                        O = Sp [j] ;
                        nzy = Sp [j+1] - O ;
                        S [k] = SVMsparse_dot (Sxi, Sii, nzx, Sx+O, Si+O, nzy) ;
                        k++ ;
                    }
                }
            }
            else if ( kernel == 1 )
            {
                for (i = 0; i < nsub; i++)
                {
                    if ( kp == i ) /* copy elements from Sold */
                    {
                        SVMfloat_scatter (S+k, SPold, kput, knum) ;
                        SPold = SPold+knum ;
                        knum-- ;
                        kput = kput+1 ;
                        kp = *kput ;
                        jp = kp ;
                        jput = kput ;
                        k += i + 1 ;
                    }
                    else
                    {
                        k += i ;
                        t = gamma*NormAx [i] ;
                        if ( degree == 3 ) t = t*t*t ;
                        else if ( degree == 2 )   t = t*t ;
                        else if ( degree == 4 ) { t = t*t; t = t*t ; }
                        else                      t = pow(t, degree) ;
                        S [k] = t ;
                        k++ ;
                    }
                    nzx = Sp [i+1] - Sp [i] ;
                    Sii = Si+Sp [i] ;
                    Sxi = Sx+Sp [i] ;
                    for (j = i+1; j < nsub; j++)
                    {
                        if ( j == jp ) /* this element stored in scatter above*/
                        {
                            jput = jput+1 ;
                            jp = *jput ;
                            k++ ;
                            continue ;
                        }
                        O = Sp [j] ;
                        nzy = Sp [j+1] - O ;
                        t = gamma*SVMsparse_dot(Sxi, Sii, nzx, Sx+O, Si+O, nzy);
                        if ( degree == 3 ) t = t*t*t ;
                        else if ( degree == 2 )   t = t*t ;
                        else if ( degree == 4 ) { t = t*t; t = t*t ; }
                        else                      t = pow(t, degree) ;
                        S [k] = t ;
                        k++ ;
                    }
                }
            }
            else if ( kernel == 2 )
            {
                for (i = 0; i < nsub; i++)
                {
                    sx = NormAx [i] ;
                    if ( kp == i ) /* copy elements from Sold */
                    {
                        SVMfloat_scatter (S+k, SPold, kput, knum) ;
                        SPold = SPold+knum ;
                        knum-- ;
                        kput = kput+1 ;
                        kp = *kput ;
                        jp = kp ;
                        jput = kput ;
                        k += i + 1 ;
                    }
                    else
                    {
                        k += i ;
                        S [k] = 1 ;
                        k++ ;
                    }
                    nzx = Sp [i+1] - Sp [i] ;
                    Sii = Si+Sp [i] ;
                    Sxi = Sx+Sp [i] ;
                    for (j = i+1; j < nsub; j++)
                    {
                        if ( j == jp ) /* this element stored in scatter above*/
                        {
                            jput = jput+1 ;
                            jp = *jput ;
                            k++ ;
                            continue ;
                        }
                        O = Sp [j] ;
                        nzy = Sp [j+1] - O ;
                        t = SVMsparse_dot (Sxi, Sii, nzx, Sx+O, Si+O, nzy) ;
                        S [k] = exp (gamma*(2.*t - sx - NormAx [j])) ;
                        k++ ;
                    }
                }
            }
            else /* kernel == 3 */
            {
                for (i = 0; i < nsub; i++)
                {
                    if ( kp == i ) /* copy elements from Sold */
                    {
                        SVMfloat_scatter (S+k, SPold, kput, knum) ;
                        SPold = SPold+knum ;
                        knum-- ;
                        kput = kput+1 ;
                        kp = *kput ;
                        jp = kp ;
                        jput = kput ;
                        k += i + 1 ;
                    }
                    else
                    {
                        k += i ;
                        S [k] = tanh (gamma*(double) NormAx [i]) ;
                        k++ ;
                    }
                    nzx = Sp [i+1] - Sp [i] ;
                    Sii = Si+Sp [i] ;
                    Sxi = Sx+Sp [i] ;
                    for (j = i+1; j < nsub; j++)
                    {
                        if ( j == jp ) /* this element stored in scatter above*/
                        {
                            jput = jput+1 ;
                            jp = *jput ;
                            k++ ;
                            continue ;
                        }
                        O = Sp [j] ;
                        nzy = Sp [j+1] - O ;
                        t = SVMsparse_dot (Sxi, Sii, nzx, Sx+O, Si+O, nzy) ;
                        S [k] = tanh (gamma*t) ;
                        k++ ;
                    }
                }
            }
        }
        else /* not SparseDot */
        {
            if ( kernel == 0 )
            {
                for (i = 0; i < nsub; i++)
                {
                    if ( kp == i ) /* copy elements from Sold */
                    {
                        SVMfloat_scatter (S+k, SPold, kput, knum) ;
                        SPold = SPold+knum ;
                        knum-- ;
                        kput = kput+1 ;
                        kp = *kput ;
                        jp = kp ;
                        jput = kput ;
                        k += i + 1 ;
                    }
                    else
                    {
                        k += i ;
                        S [k] = NormAx [i] ;
                        k++ ;
                    }
                    l = Sp [i+1] ;
                    for (j = Sp [i]; j < l; j++) Colx [Si [j]] = Sx [j]  ;
                    for (j = i+1; j < nsub; j++)
                    {
                        if ( j == jp ) /* this element stored in scatter above*/
                        {
                            jput = jput+1 ;
                            jp = *jput ;
                            k++ ;
                            continue ;
                        }
                        O = Sp [j] ;
                        nzy = Sp [j+1] - O ;
                        S [k] = SVMfloat_dense_dot (Colx, Sx+O, Si+O, nzy) ;
                        k++ ;
                    }
                    for (j = Sp [i]; j < l; j++) Colx [Si [j]] = 0. ;
                }
            }
            else if ( kernel == 1 )
            {
                for (i = 0; i < nsub; i++)
                {
                    if ( kp == i ) /* copy elements from Sold */
                    {
                        SVMfloat_scatter (S+k, SPold, kput, knum) ;
                        SPold = SPold+knum ;
                        knum-- ;
                        kput = kput+1 ;
                        kp = *kput ;
                        jp = kp ;
                        jput = kput ;
                        k += i + 1 ;
                    }
                    else
                    {
                        k += i ;
                        t = gamma*NormAx [i] ;
                        if ( degree == 3 ) t = t*t*t ;
                        else if ( degree == 2 )   t = t*t ;
                        else if ( degree == 4 ) { t = t*t; t = t*t ; }
                        else                      t = pow(t, degree) ;
                        S [k] = t ;
                        k++ ;
                    }
                    l = Sp [i+1] ;
                    for (j = Sp [i]; j < l; j++) Colx [Si [j]] = Sx [j] ;
                    for (j = i+1; j < nsub; j++)
                    {
                        if ( j == jp ) /* this element stored in scatter above*/
                        {
                            jput = jput+1 ;
                            jp = *jput ;
                            k++ ;
                            continue ;
                        }
                        O = Sp [j] ;
                        nzy = Sp [j+1] - O ;
                        t = gamma*SVMfloat_dense_dot (Colx, Sx+O, Si+O, nzy) ;
                        if ( degree == 3 ) t = t*t*t ;
                        else if ( degree == 2 )   t = t*t ;
                        else if ( degree == 4 ) { t = t*t; t = t*t ; }
                        else                      t = pow(t, degree) ;
                        S [k] = t ;
                        k++ ;
                    }
                    for (j = Sp [i]; j < l; j++) Colx [Si [j]] = 0. ;
                }
            }
            else if ( kernel == 2 )
            {
                for (i = 0; i < nsub; i++)
                {
                    sx = NormAx [i] ;
                    if ( kp == i ) /* copy elements from Sold */
                    {
                        SVMfloat_scatter (S+k, SPold, kput, knum) ;
                        SPold = SPold+knum ;
                        knum-- ;
                        kput = kput+1 ;
                        kp = *kput ;
                        jp = kp ;
                        jput = kput ;
                        k += i + 1 ;
                    }
                    else
                    {
                        k += i ;
                        S [k] = 1 ;
                        k++ ;
                    }
                    l = Sp [i+1] ;
                    for (j = Sp [i]; j < l; j++) Colx [Si [j]] = Sx [j] ;
                    for (j = i+1; j < nsub; j++)
                    {
                        if ( j == jp ) /* this element stored in scatter above*/
                        {
                            jput = jput+1 ;
                            jp = *jput ;
                            k++ ;
                            continue ;
                        }
                        O = Sp [j] ;
                        nzy = Sp [j+1] - O ;
                        t = SVMfloat_dense_dot (Colx, Sx+O, Si+O, nzy) ;
                        S [k] = exp (gamma*(2.*t - sx - NormAx [j])) ;
                        k++ ;
                    }
                    for (j = Sp [i]; j < l; j++) Colx [Si [j]] = 0. ;
                }
            }
            else /* kernel == 3 */
            {
                for (i = 0; i < nsub; i++)
                {
                    if ( kp == i ) /* copy elements from Sold */
                    {
                        SVMfloat_scatter (S+k, SPold, kput, knum) ;
                        SPold = SPold+knum ;
                        knum-- ;
                        kput = kput+1 ;
                        kp = *kput ;
                        jp = kp ;
                        jput = kput ;
                        k += i + 1 ;
                    }
                    else
                    {
                        k += i ;
                        S [k] = tanh (gamma*NormAx [i]) ;
                        k++ ;
                    }
                    l = Sp [i+1] ;
                    for (j = Sp [i]; j < l; j++) Colx [Si [j]] = Sx [j] ;
                    for (j = i+1; j < nsub; j++)
                    {
                        if ( j == jp ) /* this element stored in scatter above*/
                        {
                            jput = jput+1 ;
                            jp = *jput ;
                            k++ ;
                            continue ;
                        }
                        O = Sp [j] ;
                        nzy = Sp [j+1] - O ;
                        t = SVMfloat_dense_dot (Colx, Sx+O, Si+O, nzy) ;
                        S [k] = tanh (gamma*t) ;
                        k++ ;
                    }
                    for (j = Sp [i]; j < l; j++) Colx [Si [j]] = 0. ;
                }
            }
        }
    }
    /* Scatter the symmetric part of S (copy into rows from columns).
       S should be small enough to fit in cache for this to be efficient */
    k = nsub ;
    for (i = 1; i < nsub; i++)
    {
        l = i ;
        for (j = 0; j < i; j++)
        {
            S [k] = S [l] ;
            k++ ;
            l += nsub ;
        }
        k += nsub - i ;
    }
}

PRIVATE void SVMupdateGrad
(
    double       *g, /* gradient to be updated */
    double       *d, /* change in the subspace: gnew = g + A*d */
    double       *y, /* solution to subspace problem */
    int          *I, /* indices of columns in subspace */
    int        nsub, /* subspace size */
    int     *istore, /* indices of already stored columns for subspace */
    int      nstore, /* number of already stored columns for subspace */
    int      *ifree, /* indices for free variables in subspace */
    int      *nfree, /* number free variables in prior subspace (input)
                        number free variables in new subspace (output) */
    float        *S, /* subspace matrix */
    float       *Sx, /* numerical data, compressed in routine */
    short       *Si, /* features associated with data, compressed in routine */
    int         *Sp, /* pointer array, compressed in routine */
    float     *Colx, /* storage for float column of data */
    short     *Coli, /* storage for 1s column of data */
    double  *NormAx, /* norm of subspace columns, compressed in routine */
    short   *NormAi, /* norm of subspace columns, compressed in routine */
    SVMcom     *Com  /* SVM Com structure */
)
{
    short Ax_is_1, degree, flag, iflag, kernel, SparseDot, SparseProduct,
         *Ai, *Aii, *normAi ;
    int bsub, i, istart, iend, inc, j, k, kg, l, m, n, nk, np, nfeature,
        Nstore, Nfree, NewFree, nzx, nzy, ncols, p, si, NewInc, OldInc, O, P,
        PrintLevel, Sinc, *Ap, *kget, *kgetS, *kput, *pget, *pput,
       *zi, *OldCols ;
    float *Anew, *Aold, *Ax, *Axi, *SP, *A ;
    double lo_pert, hi_pert, gamma, r, t, sx, *normAx, *LinWork, *z ;
    SVMparm *Parm ;

    /* ======== extract variables ======== */
    kernel = Com->kernel ;
    bsub = Com->bsub ;
    nfeature = Com->nfeature ;
    normAi = Com->normAi ;
    normAx = Com->normAx ;
    if ( kernel > 0 ) /* columns are saved */
    {
        A = Com->Astore ;
        z = Com->xWorkUpdate ;
        pput = Com->iWorkUpdate ;
        pget = pput+bsub ;
        kput = pget+bsub ;
        kget = kput+bsub ;
        kgetS= kget+(bsub+1) ;
        zi = kgetS+bsub ;
        OldCols = zi+bsub ;
    }
    else LinWork = Com->xWorkUpdate ;

    Parm = Com->Parm ;
    PrintLevel = Parm->PrintLevel ;
    degree = Parm->degree ;
    gamma = Parm->gamma ;
    Ax_is_1 = Parm->Ax_is_1 ;
    SparseDot  = Parm->SparseDot  ;
    Ax = Com->Ax ;
    Ai = Com->Ai ;
    Ap = Com->Ap ;
    n = Com->n ;
    lo_pert = Com->C*Parm->bnd_pert ;
    hi_pert = Com->C - lo_pert ;
    /* ======== end of extract ======== */

    I [nsub] = n ; /* flag the end of the array of subspace indices */

    if ( kernel == 0 ) /* !!! the kernel is linear, no columns saved !!! */
    {
        SVMscalar_init (LinWork, 0., nfeature) ;
        if ( Ax_is_1 )
        {
            for (j = 0; j < nsub; j++)
            {
                if ( d [j] != 0. )
                {
                    k = Sp [j] ;
                    SVMsparse_add (LinWork, Si+k, d [j], Sp [j+1] - k) ;
                }
            }
            k = 0 ;
            for (j = 0; j < n; j++)
            {
                if ( I [k] == j ) { k++; continue ; }/* computed in main prog */
                l = Ap [j] ;
                g [j] += SVMbinary_double_dot (LinWork, Ai+l, Ap [j+1] - l) ;
            }
        }
        else
        {
            for (j = 0; j < nsub; j++)
            {
                if ( d [j] != 0. )
                {
                    k = Sp [j] ;
                    SVMsparse_daxpy (LinWork, Sx+k, Si+k, d [j], Sp [j+1] - k) ;
                }
            }
            k = 0 ;
            for (j = 0; j < n; j++)
            {
                if ( I [k] == j ) { k++; continue ; }/* computed in main prog */
                l = Ap [j] ;
                g [j] += SVMdouble_dense_dot (LinWork, Ax+l, Ai+l, Ap [j+1]-l) ;
            }
        }

        /* determine the new free indices */

        Nfree = 0 ;
        for (j = 0; j < nsub; j++)
        {
            if ( (y [j] > lo_pert) && (y [j] < hi_pert) )
            {
                ifree [Nfree] = I [j] ;
                Nfree++ ;
            }
        }
        *nfree = Nfree ;
        if ( PrintLevel > 1 )
            printf ("    free indices:           %i\n", Nfree) ;
        return ;
    }

    /* !!! the kernel is nonlinear, columns are saved !!! */

    /* ifree initially contains the free indices from the prior subspace.
       These indices also correspond to the columns that were saved in
       the previous iteration. istore contains the indices of the saved
       columns that are also indices for the current subspace. Hence,
       we will utilize these columns when we evaluate the gradient change */

    Nfree = *nfree ;       /* prior number of freed variables = # saved cols */
    Nstore = nstore ;      /* number of prior freed that will be used */
    ncols = nsub - nstore ;/* number of columns in compressed S matrix */
    if ( PrintLevel > 1 ) printf ("    repeat cols (subspace)  %i\n", nstore) ;

    /* Compress Sx array by removing the istore columns, the indices
       in istore are stored columns so we do not have to compute them. */
    if ( Nstore > Nfree/3 ) /* use dense dot product */
    {
        SparseProduct = SVMFALSE ;
        SVMscalar_init (z, SVMZERO, Nfree) ;
    }
    else SparseProduct = SVMTRUE ;
    P = 0 ; /* accumulates number of elements in Ax for skipped columns */
    k = 0 ; /* location of a column in A storeage */
    l = 0 ; /* location in current subspace */
/*printf ("use %i columns from Astore\n", nstore) ;*/
/*print_arrayi (istore, "istore", nstore) ;
print_arrayi (istore, "ifree", *nfree) ;*/
    
    for (j = 0; j < nstore; j++)
    {
        /* previously saved column in istore */
        i = istore [j] ;

        /* all the previously free columns were saved, find the location of i
           in the list of previously free columns */
        m = Sp [l] ;
        while ( i > I [l] )
        {
            /* compress d and Sp */
            Sp [l-j] = Sp [l] - P ;
            d [l-j] = d [l] ;
            if      ( NormAi != NULL ) NormAi [l-j] = NormAi [l] ;
            else if ( NormAx != NULL ) NormAx [l-j] = NormAx [l] ;
            l++ ;
        }

        /* compress Sx */
        if ( !Ax_is_1 ) SVMfloatCopy (Sx+(m-P), Sx+m, Sp [l]-m) ;

        /* compress Si */
        SVMshortCopy (Si+(m-P), Si+m, Sp [l]-m) ;

        while ( i > ifree [k] ) k++ ; /* find location in A storeage */
        if ( SparseProduct )
        {
            z [j] = d [l] ;
            zi [j] = k ;
        }
        else z [k] = d [l] ;

        P += Sp [l+1] - Sp [l] ;
        l++ ;
    }
    m = Sp [l] ;
    for (; l < nsub; l++)
    {
        Sp [l-j] = Sp [l] - P ;
        d [l-j] = d [l] ;
        if      ( NormAi != NULL ) NormAi [l-j] = NormAi [l] ;
        else if ( NormAx != NULL ) NormAx [l-j] = NormAx [l] ;
    }
    Sp [l-j] = Sp [l] - P ;
    if ( !Ax_is_1 ) SVMfloatCopy (Sx+(m-P), Sx+m, Sp [l]-m) ;
    SVMshortCopy (Si+(m-P), Si+m, Sp [l]-m) ;

    /* determine the new free indices, and partition them into
       a preserved set (OldCols that will be retained) and
       a keep set (free columns that will be evaluated and saved)
       pget = location in row of storage of a preserved column
       pput = target location in row of this preserved column
       kget = location in the subspace of the column to be kept
       kput = target location in row of the column to be kept */

    /* Save copy of ifree, the stored columns of A match ifree */
    SVMintCopy (OldCols, ifree, Nfree) ;

    OldCols [Nfree] = n ;
    istore [nstore] = n ;
    NewFree = 0 ;
    nk = 0 ;
    np = 0 ;
    l = 0 ;
    m = 0 ;
    p = 0 ;
    /* determine new free variables */
    for (j = 0; j < nsub; j++)
    {
        k = I [j] ;
        if ( (y [j] > lo_pert) && (y [j] < hi_pert) )
        {
            ifree [NewFree] = k ;
            while ( OldCols [l] < k ) l++ ;
            if ( OldCols [l] == k ) /* preserve it */
            {
                pget [np] = l ;       /* present location in row */
                pput [np] = NewFree ; /* target location in row of new A */
                np++ ;
            }
            else                      /* evaluate and save it */
            {
                kget [nk] = p ;       /* location in compressed subspace */
                kgetS[nk] = j ;       /* location in uncompressed subspace */
                kput [nk] = NewFree ; /* target location in row of new A */
                nk++ ;
            }
            NewFree++ ;
        }
        if ( istore [m] == k ) m++ ;
        else                   p++ ; /* p is location in compressed subspace */
    }
    kget [nk] = n ; /* flag at end of array */
    if ( PrintLevel > 1 )
    {
        printf ("    free indices:           %i\n", NewFree) ;
        printf ("    --retrained free:       %i\n", np) ;
        printf ("    --new free:             %i\n", nk) ;
    }

    if ( NewFree <= Nfree ) /* compress storage */
    {
/*printf ("compress storage\n") ;*/
        Aold = A ;         /* work from the start of A */
        Anew = A ;
        SP = S ;
        OldInc = Nfree ;
        NewInc = NewFree ;
        Sinc = nsub ;
        istart = 0 ;
        iend = n ;
        inc = 1 ;
        k = 0 ;
        /* when the loop index i reaches iflag, then we do not need to
           worry about collisions when copying Aold into Anew */
        flag = -1 ;
        if ( NewFree == Nfree ) iflag =-1 ;
        else                    iflag = 1 + NewFree/(Nfree - NewFree) ;
    }
    else                         /* expand storage */
    {
/*printf ("expand storage\n") ;*/
        Aold = A+(Nfree*(n-1)) ; /* work from the end of A */
        Anew = A+(NewFree*(n-1)) ;
        SP = S+(nsub*nsub - nsub) ;
        OldInc = -Nfree ;
        NewInc = -NewFree ;
        Sinc = -nsub ;
        istart = n-1 ;
        iend = -1 ;
        inc = -1 ;
        k = nsub - 1 ;
        /* when the loop index i reaches iflag, then collisions
           when copying Aold into Anew are possible */
        flag = 0 ;
        iflag = SVMMIN (1 + Nfree/(NewFree-Nfree), n-1) ;
    }
    *nfree = NewFree ;
/*printf ("SparseDot: %i SparseProduct: %i\n", SparseDot, SparseProduct) ;
printf ("OldInc: %i NewInc: %i\n", OldInc, NewInc) ;*/

    /* nonlinear kernels */
    if ( Ax_is_1 )
    {
        if ( SparseDot )
        {
            if ( kernel == 1 )
            {
                for (i = istart; i != iend; i += inc)
                {
                    /* compute product with saved columns of A */
                    if ( SparseProduct )
                    {
                        r = SVMdense_sparse_dot (z, Aold, zi, Nstore) ;
                    }
                    else r = SVMfloat_double_dot (Aold, z, Nfree) ;

                    /* copy the preserved columns */
                    if ( i == iflag )
                    {
                        if ( inc > 0 ) flag = 0 ; /* no collisions possible */
                        else           flag = 1 ; /* collisions possible */
                    }
                    SVMindexCopyFlag (Anew, Aold, pget, pput, np, flag) ;

                    /* check if index in the subspace */
                    if ( I [k] == i )
                    {
                        /* copy elements of S */
                        SVMindexCopy (Anew, SP, kgetS, kput, nk) ;

                        /* advance to the next row */
                        k += inc ;
                        SP = SP+Sinc ;
                        Anew = Anew+NewInc ;
                        Aold = Aold+OldInc ;
                        continue ; /* gradient already updated in main prog */
                    }

                    /* continue to update gradient by evaluating columns */
                    nzx = Ap [i+1] - Ap [i] ;
                    Aii = Ai+Ap [i] ;
                    m = 0 ;
                    kg = kget [m] ;
                    for (j = 0; j < ncols; j++)
                    {
                        O = Sp [j] ;
                        nzy = Sp [j+1] - O ;
                        t = gamma*SVMbinary_sparse_dot (Aii, nzx, Si+O, nzy) ;
                        if ( degree == 3 ) t = t*t*t ;
                        else if ( degree == 2 )   t = t*t ;
                        else if ( degree == 4 ) { t = t*t; t = t*t ; }
                        else                      t = pow(t, degree) ;
                        r += t*d [j] ;
                        if ( j == kg ) /* save free column */
                        {
                            Anew [kput [m]] = t ;
                            m++ ;
                            kg = kget [m] ;
                        }
                    }
                    g [i] += r ;
                    Anew = Anew+NewInc ;
                    Aold = Aold+OldInc ;
                }
            }
            else if ( kernel == 2 )
            {
                for (i = istart; i != iend; i += inc)
                {
                    /* compute product with saved columns of A */
                    if ( SparseProduct )
                    {
                        r = SVMdense_sparse_dot(z, Aold, zi, Nstore) ;
                    }
                    else r = SVMfloat_double_dot(Aold, z, Nfree) ;

                    /* copy the preserved columns */
                    if ( i == iflag )
                    {
                        if ( inc > 0 ) flag = 0 ;
                        else           flag = 1 ;
                    }
                    SVMindexCopyFlag (Anew, Aold, pget, pput, np, flag) ;

                    /* check if index in the subspace */
                    if ( I [k] == i )
                    {
                        /* copy elements of S */
                        SVMindexCopy (Anew, SP, kgetS, kput, nk) ;

                        k += inc ;
                        SP = SP+Sinc ;
                        Anew = Anew+NewInc ;
                        Aold = Aold+OldInc ;
                        continue ; /* gradient already updated in main prog */
                    }

                    /* continue to update gradient by evaluating columns */
                    si = normAi [i] ;
                    nzx = Ap [i+1] - Ap [i] ;
                    Aii = Ai+Ap [i] ;
                    m = 0 ;
                    kg = kget [m] ;
                    for (j = 0; j < ncols; j++)
                    {
                        O = Sp [j] ;
                        nzy = Sp [j+1] - O ;
                        t = SVMbinary_sparse_dot (Aii, nzx, Si+O, nzy) ;
                        t = exp (gamma*(2.*t - si - NormAi [j])) ;
                        r += t*d [j] ;
                        if ( j == kg ) /* save free column */
                        {
                            Anew [kput [m]] = t ;
                            m++ ;
                            kg = kget [m] ;
                        }
                    }
                    g [i] += r ;
                    Anew = Anew+NewInc ;
                    Aold = Aold+OldInc ;
                }
            }
            else /* kernel == 3 */
            {
                for (i = istart; i != iend; i += inc)
                {
                    /* compute product with saved columns of A */
                    if ( SparseProduct )
                    {
                        r = SVMdense_sparse_dot(z, Aold, zi, Nstore) ;
                    }
                    else r = SVMfloat_double_dot(Aold, z, Nfree) ;

                    /* copy the preserved columns */
                    if ( i == iflag )
                    {
                        if ( inc > 0 ) flag = 0 ;
                        else           flag = 1 ;
                    }
                    SVMindexCopyFlag (Anew, Aold, pget, pput, np, flag) ;

                    /* check if index in the subspace */
                    if ( I [k] == i )
                    {
                        /* copy elements of S */
                        SVMindexCopy (Anew, SP, kgetS, kput, nk) ;

                        /* advance to the next row */
                        k += inc ;
                        SP = SP+Sinc ;
                        Anew = Anew+NewInc ;
                        Aold = Aold+OldInc ;
                        continue ; /* gradient already updated in main prog */
                    }

                    /* continue to update gradient by evaluating columns */
                    nzx = Ap [i+1] - Ap [i] ;
                    Aii = Ai+Ap [i] ;
                    m = 0 ;
                    kg = kget [m] ;
                    for (j = 0; j < ncols; j++)
                    {
                        O = Sp [j] ;
                        nzy = Sp [j+1] - O ;
                        t = SVMbinary_sparse_dot (Aii, nzx, Si+O, nzy) ;
                        t = tanh (gamma*t) ;
                        r += t*d [j] ;
                        if ( j == kg ) /* save free column */
                        {
                            Anew [kput [m]] = t ;
                            m++ ;
                            kg = kget [m] ;
                        }
                    }
                    g [i] += r ;
                    Anew = Anew+NewInc ;
                    Aold = Aold+OldInc ;
                }
            }
        }
        else /* not SparseDot */
        {
            if ( kernel == 1 )
            {
                for (i = istart; i != iend; i += inc)
                {
                    /* compute product with saved columns of A */
                    if ( SparseProduct )
                    {
                        r = SVMdense_sparse_dot (z, Aold, zi, Nstore) ;
                    }
                    else r = SVMfloat_double_dot (Aold, z, Nfree) ;

                    /* copy the preserved columns */
                    if ( i == iflag )
                    {
                        if ( inc > 0 ) flag = 0 ;
                        else           flag = 1 ;
                    }
                    SVMindexCopyFlag (Anew, Aold, pget, pput, np, flag) ;

                    /* check if index in the subspace */
                    if ( I [k] == i )
                    {
                        /* copy elements of S */
                        SVMindexCopy (Anew, SP, kgetS, kput, nk) ;

                        /* advance to the next row */
                        k += inc ;
                        SP = SP+Sinc ;
                        Anew = Anew+NewInc ;
                        Aold = Aold+OldInc ;
                        continue ; /* gradient already updated in main prog */
                    }

                    /* continue to update gradient by evaluating columns */
                    l = Ap [i+1] ;
                    for (j = Ap [i]; j < l; j++) Coli [Ai [j]] = 1 ;
                    m = 0 ;
                    kg = kget [m] ;
                    for (j = 0; j < ncols; j++)
                    {
                        O = Sp [j] ;
                        nzy = Sp [j+1] - O ;
                        t = gamma*SVMbinary_short_dot (Coli, Si+O, nzy) ;
                        if ( degree == 3 ) t = t*t*t ;
                        else if ( degree == 2 )   t = t*t ;
                        else if ( degree == 4 ) { t = t*t; t = t*t ; }
                        else                      t = pow(t, degree) ;
                        r += t*d [j] ;
                        if ( j == kg ) /* save free column */
                        {
                            Anew [kput [m]] = t ;
                            m++ ;
                            kg = kget [m] ;
                        }
                    }
                    g [i] += r ;
                    for (j = Ap [i]; j < l; j++) Coli [Ai [j]] = 0 ;
                    Anew = Anew+NewInc ;
                    Aold = Aold+OldInc ;
                }
            }
            else if ( kernel == 2 )
            {
                for (i = istart; i != iend; i += inc)
                {
                    /* compute product with saved columns of A */
                    if ( SparseProduct )
                    {
                        r = SVMdense_sparse_dot (z, Aold, zi, Nstore) ;
                    }
                    else r = SVMfloat_double_dot (Aold, z, Nfree) ;

                    /* copy the preserved columns */
                    if ( i == iflag )
                    {
                        if ( inc > 0 ) flag = 0 ;
                        else           flag = 1 ;
                    }
                    SVMindexCopyFlag (Anew, Aold, pget, pput, np, flag) ;

                    /* check if index in the subspace */
                    if ( I [k] == i )
                    {
                        /* copy elements of S */
                        SVMindexCopy (Anew, SP, kgetS, kput, nk) ;

                        /* advance to the next row */
                        k += inc ;
                        SP = SP+Sinc ;
                        Anew = Anew+NewInc ;
                        Aold = Aold+OldInc ;
                        continue ; /* gradient already updated in main prog */
                    }

                    /* continue to update gradient by evaluating columns */
                    l = Ap [i+1] ;
                    for (j = Ap [i]; j < l; j++) Coli [Ai [j]] = 1 ;
                    si = normAi [i] ;
                    m = 0 ;
                    kg = kget [m] ;
                    for (j = 0; j < ncols; j++)
                    {
                        O = Sp [j] ;
                        nzy = Sp [j+1] - O ;
                        t = SVMbinary_short_dot (Coli, Si+O, nzy) ;
                        t = exp (gamma*(2.*t - si - NormAi [j])) ;
                        r += t*d [j] ;
                        if ( j == kg ) /* save free column */
                        {
                            Anew [kput [m]] = t ;
                            m++ ;
                            kg = kget [m] ;
                        }
                    }
                    g [i] += r ;
                    for (j = Ap [i]; j < l; j++) Coli [Ai [j]] = 0 ;
                    Anew = Anew+NewInc ;
                    Aold = Aold+OldInc ;
                }
            }
            else /* kernel == 3 */
            {
                for (i = istart; i != iend; i += inc)
                {
                    /* compute product with saved columns of A */
                    if ( SparseProduct )
                    {
                        r = SVMdense_sparse_dot (z, Aold, zi, Nstore) ;
                    }
                    else r = SVMfloat_double_dot (Aold, z, Nfree) ;

                    /* copy the preserved columns */
                    if ( i == iflag )
                    {
                        if ( inc > 0 ) flag = 0 ;
                        else           flag = 1 ;
                    }
                    SVMindexCopyFlag (Anew, Aold, pget, pput, np, flag) ;

                    /* check if index in the subspace */
                    if ( I [k] == i )
                    {
                        /* copy elements of S */
                        SVMindexCopy (Anew, SP, kgetS, kput, nk) ;

                        /* advance to the next row */
                        k += inc ;
                        SP = SP+Sinc ;
                        Anew = Anew+NewInc ;
                        Aold = Aold+OldInc ;
                        continue ; /* gradient already updated in main prog */
                    }

                    /* continue to update gradient by evaluating columns */
                    l = Ap [i+1] ;
                    for (j = Ap [i]; j < l; j++) Coli [Ai [j]] = 1 ;
                    m = 0 ;
                    kg = kget [m] ;
                    for (j = 0; j < ncols; j++)
                    {
                        O = Sp [j] ;
                        nzy = Sp [j+1] - O ;
                        t = SVMbinary_short_dot (Coli, Si+O, nzy) ;
                        t = tanh (gamma*t) ;
                        r += t*d [j] ;
                        if ( j == kg ) /* save free column */
                        {
                            Anew [kput [m]] = t ;
                            m++ ;
                            kg = kget [m] ;
                        }
                    }
                    g [i] += r ;
                    for (j = Ap [i]; j < l; j++) Coli [Ai [j]] = 0 ;
                    Anew = Anew+NewInc ;
                    Aold = Aold+OldInc ;
                }
            }
        }
    }
    else /* Ax != 1 */
    {
        if ( SparseDot )
        {
            if ( kernel == 1 )
            {
                for (i = istart; i != iend; i += inc)
                {
                    /* compute product with saved columns of A */
                    if ( SparseProduct )
                    {
                        r = SVMdense_sparse_dot(z, Aold, zi, Nstore) ;
                    }
                    else r = SVMfloat_double_dot(Aold, z, Nfree) ;

                    /* copy the preserved columns */
                    if ( i == iflag )
                    {
                        if ( inc > 0 ) flag = 0 ;
                        else           flag = 1 ;
                    }
                    SVMindexCopyFlag (Anew, Aold, pget, pput, np, flag) ;

                    /* check if index in the subspace */
                    if ( I [k] == i )
                    {
                        /* copy elements of S */
                        SVMindexCopy (Anew, SP, kgetS, kput, nk) ;

                        /* advance to the next row */
                        k += inc ;
                        SP = SP+Sinc ;
                        Anew = Anew+NewInc ;
                        Aold = Aold+OldInc ;
                        continue ; /* gradient already updated in main prog */
                    }

                    /* continue to update gradient by evaluating columns */
                    nzx = Ap [i+1] - Ap [i] ;
                    Aii = Ai+Ap [i] ;
                    Axi = Ax+Ap [i] ;
                    m = 0 ;
                    kg = kget [m] ;
                    for (j = 0; j < ncols; j++)
                    {
                        O = Sp [j] ;
                        nzy = Sp [j+1] - O ;
                        t = gamma*SVMsparse_dot(Axi, Aii, nzx, Sx+O, Si+O, nzy);
                        if ( degree == 3 ) t = t*t*t ;
                        else if ( degree == 2 )   t = t*t ;
                        else if ( degree == 4 ) { t = t*t; t = t*t ; }
                        else                      t = pow(t, degree) ;
                        r += t*d [j] ;
                        if ( j == kg ) /* save free column */
                        {
                            Anew [kput [m]] = t ;
                            m++ ;
                            kg = kget [m] ;
                        }
                    }
                    g [i] += r ;
                    Anew = Anew+NewInc ;
                    Aold = Aold+OldInc ;
                }
            }
            else if ( kernel == 2 )
            {
                for (i = istart; i != iend; i += inc)
                {
                    /* compute product with saved columns of A */
                    if ( SparseProduct )
                    {
                        r = SVMdense_sparse_dot(z, Aold, zi, Nstore) ;
                    }
                    else r = SVMfloat_double_dot(Aold, z, Nfree) ;

                    /* copy the preserved columns */
                    if ( i == iflag )
                    {
                        if ( inc > 0 ) flag = 0 ;
                        else           flag = 1 ;
                    }
                    SVMindexCopyFlag (Anew, Aold, pget, pput, np, flag) ;

                    /* check if index in the subspace */
                    if ( I [k] == i )
                    {
                        /* copy elements of S */
                        SVMindexCopy (Anew, SP, kgetS, kput, nk) ;

                        /* advance to the next row */
                        k += inc ;
                        SP = SP+Sinc ;
                        Anew = Anew+NewInc ;
                        Aold = Aold+OldInc ;
                        continue ; /* gradient already updated in main prog */
                    }

                    /* continue to update gradient by evaluating columns */
                    sx = normAx [i] ;
                    nzx = Ap [i+1] - Ap [i] ;
                    Aii = Ai+Ap [i] ;
                    Axi = Ax+Ap [i] ;
                    m = 0 ;
                    kg = kget [m] ;
                    for (j = 0; j < ncols; j++)
                    {
                        O = Sp [j] ;
                        nzy = Sp [j+1] - O ;
                        t = SVMsparse_dot (Axi, Aii, nzx, Sx+O, Si+O, nzy) ;
                        t = exp (gamma*(2.*t - sx - NormAx [j])) ;
                        r += t*d [j] ;
                        if ( j == kg ) /* save free column */
                        {
                            Anew [kput [m]] = t ;
                            m++ ;
                            kg = kget [m] ;
                        }
                    }
                    g [i] += r ;
                    Anew = Anew+NewInc ;
                    Aold = Aold+OldInc ;
                }
            }
            else /* kernel == 3 */
            {
                for (i = istart; i != iend; i += inc)
                {
                    /* compute product with saved columns of A */
                    if ( SparseProduct )
                    {
                        r = SVMdense_sparse_dot(z, Aold, zi, Nstore) ;
                    }
                    else r = SVMfloat_double_dot(Aold, z, Nfree) ;

                    /* copy the preserved columns */
                    if ( i == iflag )
                    {
                        if ( inc > 0 ) flag = 0 ;
                        else           flag = 1 ;
                    }
                    SVMindexCopyFlag (Anew, Aold, pget, pput, np, flag) ;

                    /* check if index in the subspace */
                    if ( I [k] == i )
                    {
                        /* copy elements of S */
                        SVMindexCopy (Anew, SP, kgetS, kput, nk) ;

                        /* advance to the next row */
                        k += inc ;
                        SP = SP+Sinc ;
                        Anew = Anew+NewInc ;
                        Aold = Aold+OldInc ;
                        continue ; /* gradient already updated in main prog */
                    }

                    /* continue to update gradient by evaluating columns */
                    nzx = Ap [i+1] - Ap [i] ;
                    Aii = Ai+Ap [i] ;
                    Axi = Ax+Ap [i] ;
                    m = 0 ;
                    kg = kget [m] ;
                    for (j = 0; j < ncols; j++)
                    {
                        O = Sp [j] ;
                        nzy = Sp [j+1] - O ;
                        t = SVMsparse_dot (Axi, Aii, nzx, Sx+O, Si+O, nzy) ;
                        t = tanh (gamma*t) ;
                        r += t*d [j] ;
                        if ( j == kg ) /* save free column */
                        {
                            Anew [kput [m]] = t ;
                            m++ ;
                            kg = kget [m] ;
                        }
                    }
                    g [i] += r ;
                    Anew = Anew+NewInc ;
                    Aold = Aold+OldInc ;
                }
            }
        }
        else /* not SparseDot */
        {
            if ( kernel == 1 )
            {
                for (i = istart; i != iend; i += inc)
                {
                    /* compute product with saved columns of A */
                    if ( SparseProduct )
                    {
                        r = SVMdense_sparse_dot(z, Aold, zi, Nstore) ;
                    }
                    else r = SVMfloat_double_dot(Aold, z, Nfree) ;

                    /* copy the preserved columns */
                    if ( i == iflag )
                    {
                        if ( inc > 0 ) flag = 0 ;
                        else           flag = 1 ;
                    }
                    SVMindexCopyFlag (Anew, Aold, pget, pput, np, flag) ;

                    /* check if index in the subspace */
                    if ( I [k] == i )
                    {
                        /* copy elements of S */
                        SVMindexCopy (Anew, SP, kgetS, kput, nk) ;

                        /* advance to the next row */
                        k += inc ;
                        SP = SP+Sinc ;
                        Anew = Anew+NewInc ;
                        Aold = Aold+OldInc ;
                        continue ; /* gradient already updated in main prog */
                    }

                    /* continue to update gradient by evaluating columns */
                    l = Ap [i+1] ;
                    for (j = Ap [i]; j < l; j++) Colx [Ai [j]] = Ax [j] ;
                    m = 0 ;
                    kg = kget [m] ;
                    for (j = 0; j < ncols; j++)
                    {
                        O = Sp [j] ;
                        nzy = Sp [j+1] - O ;
                        t = gamma*SVMfloat_dense_dot (Colx, Sx+O, Si+O, nzy) ;
                        if ( degree == 3 ) t = t*t*t ;
                        else if ( degree == 2 )   t = t*t ;
                        else if ( degree == 4 ) { t = t*t; t = t*t ; }
                        else                      t = pow(t, degree) ;
                        r += t*d [j] ;
                        if ( j == kg ) /* save free column */
                        {
                            Anew [kput [m]] = t ;
                            m++ ;
                            kg = kget [m] ;
                        }
                    }
                    g [i] += r ;
                    for (j = Ap [i]; j < l; j++) Colx [Ai [j]] = 0. ;
                    Anew = Anew+NewInc ;
                    Aold = Aold+OldInc ;
                }
            }
            else if ( kernel == 2 )
            {
                for (i = istart; i != iend; i += inc)
                {
                    /* compute product with saved columns of A */
                    if ( SparseProduct )
                    {
                        r = SVMdense_sparse_dot(z, Aold, zi, Nstore) ;
                    }
                    else r = SVMfloat_double_dot(Aold, z, Nfree) ;

                    /* copy the preserved columns */
                    if ( i == iflag )
                    {
                        if ( inc > 0 ) flag = 0 ;
                        else           flag = 1 ;
                    }
                    SVMindexCopyFlag (Anew, Aold, pget, pput, np, flag) ;

                    /* check if index in the subspace */
                    if ( I [k] == i )
                    {
                        /* copy elements of S */
                        SVMindexCopy (Anew, SP, kgetS, kput, nk) ;

                        /* advance to the next row */
                        k += inc ;
                        SP = SP+Sinc ;
                        Anew = Anew+NewInc ;
                        Aold = Aold+OldInc ;
                        continue ; /* gradient already updated in main prog */
                    }

                    /* continue to update gradient by evaluating columns */
                    sx = normAx [i] ;
                    l = Ap [i+1] ;
                    for (j = Ap [i]; j < l; j++) Colx [Ai [j]] = Ax [j] ;
                    m = 0 ;
                    kg = kget [m] ;
                    for (j = 0; j < ncols; j++)
                    {
                        O = Sp [j] ;
                        nzy = Sp [j+1] - O ;
                        t = SVMfloat_dense_dot (Colx, Sx+O, Si+O, nzy) ;
                        t = exp(gamma*(2.*t - sx - NormAx [j])) ;
                        r += t*d [j] ;
                        if ( j == kg ) /* save free column */
                        {
                            Anew [kput [m]] = t ;
                            m++ ;
                            kg = kget [m] ;
                        }
                    }
                    g [i] += r ;
                    for (j = Ap [i]; j < l; j++) Colx [Ai [j]] = 0. ;
                    Anew = Anew+NewInc ;
                    Aold = Aold+OldInc ;
                }
            }
            else /* kernel == 3 */
            {
                for (i = istart; i != iend; i += inc)
                {
                    /* compute product with saved columns of A */
                    if ( SparseProduct )
                    {
                        r = SVMdense_sparse_dot(z, Aold, zi, Nstore) ;
                    }
                    else r = SVMfloat_double_dot(Aold, z, Nfree) ;

                    /* copy the preserved columns */
                    if ( i == iflag )
                    {
                        if ( inc > 0 ) flag = 0 ;
                        else           flag = 1 ;
                    }
                    SVMindexCopyFlag (Anew, Aold, pget, pput, np, flag) ;

                    /* check if index in the subspace */
                    if ( I [k] == i )
                    {
                        /* copy elements of S */
                        SVMindexCopy (Anew, SP, kgetS, kput, nk) ;

                        /* advance to the next row */
                        k += inc ;
                        SP = SP+Sinc ;
                        Anew = Anew+NewInc ;
                        Aold = Aold+OldInc ;
                        continue ; /* gradient already updated in main prog */
                    }

                    /* continue to update gradient by evaluating columns */
                    l = Ap [i+1] ;
                    for (j = Ap [i]; j < l; j++) Colx [Ai [j]] = Ax [j] ;
                    m = 0 ;
                    kg = kget [m] ;
                    for (j = 0; j < ncols; j++)
                    {
                        O = Sp [j] ;
                        nzy = Sp [j+1] - O ;
                        t = SVMfloat_dense_dot (Colx, Sx+O, Si+O, nzy) ;
                        t = tanh (gamma*t) ;
                        r += t*d [j] ;
                        if ( j == kg ) /* save free column */
                        {
                            Anew [kput [m]] = t ;
                            m++ ;
                            kg = kget [m] ;
                        }
                    }
                    g [i] += r ;
                    for (j = Ap [i]; j < l; j++) Colx [Ai [j]] = 0. ;
                    Anew = Anew+NewInc ;
                    Aold = Aold+OldInc ;
                }
            }
        }
    }
}

/* dot product x'y where x is dense and y is sparse */
double SVMdouble_dense_dot
(
    double *x,  /* nonzero numerical values, dense format */
    float  *y,  /* nonzero numerical values */
    short *yi,  /* location in x of nonzero indices in y */
    int     n   /* number of nonzeros in y */
)
{
    int n5 ;
    register int j ;
    register double t ;
    t = 0 ;
    n5 = n % 5 ;
    for (j = 0; j < n5; j++) t += x [yi [j]]*y [j] ;
    for (; j < n; j += 5)
    {
        t += x [yi[j]]  *y[j]
          +  x [yi[j+1]]*y [j+1]
          +  x [yi[j+2]]*y [j+2]
          +  x [yi[j+3]]*y [j+3]
          +  x [yi[j+4]]*y [j+4] ;
    }
    return (t) ;
}

/* dot product x'y where x is dense and y is sparse */
double SVMdense_sparse_dot
(
    double *x,  /* nonzero numerical values, dense format */
    float  *y,  /* nonzero numerical values */
    int   *yi,  /* location in y of nonzero indices in x */
    int     n   /* number of nonzeros in y */
)
{
    int n5 ;
    register int j ;
    register double t ;
    t = 0. ;
    n5 = n % 5 ;
    for (j = 0; j < n5; j++) t += x [j]*y [yi [j]] ;
    for (; j < n; j += 5)
    {
        t += x [j]  *y[yi [j]]
          +  x [j+1]*y [yi [j+1]]
          +  x [j+2]*y [yi [j+2]]
          +  x [j+3]*y [yi [j+3]]
          +  x [j+4]*y [yi [j+4]] ;
    }
    return (t) ;
}

/* dot product x'y where x is dense and y is sparse */
double SVMfloat_dense_dot
(
    float  *x,  /* nonzero numerical values, dense format */
    float  *y,  /* nonzero numerical values */
    short *yi,  /* nonzero indices in y */
    int     n   /* number of nonzeros in y */
)
{
    int n5 ;
    register int j ;
    register double t ;
    t = 0. ;
    n5 = n % 5 ;
    for (j = 0; j < n5; j++) t += (double) (x [yi [j]]*y [j]) ;
    for (; j < n; j += 5)
    {
        t += (double) (x [yi[j]]  *y[j])
          +  (double) (x [yi[j+1]]*y [j+1])
          +  (double) (x [yi[j+2]]*y [j+2])
          +  (double) (x [yi[j+3]]*y [j+3])
          +  (double) (x [yi[j+4]]*y [j+4]) ;
    }
    return (t) ;
}

/* dot product x'y where x is dense and y is sparse */
double SVMbinary_double_dot
(
    double *x,  /* nonzero numerical values, dense format */
    short *yi,  /* nonzero indices in y */
    int     n   /* number of nonzeros in y */
)
{
    int n5 ;
    register int j ;
    register double t ;
    t = 0. ;
    n5 = n % 5 ;
    for (j = 0; j < n5; j++) t += x [yi [j]] ;
    for (; j < n; j += 5)
    {
        t += x [yi[j]]
          +  x [yi[j+1]]
          +  x [yi[j+2]]
          +  x [yi[j+3]]
          +  x [yi[j+4]] ;
    }
    return (t) ;
}

/* dot product x'y where x is dense and y is sparse */
short SVMbinary_short_dot
(
    short  *x,  /* nonzero numerical values, dense format */
    short *yi,  /* nonzero indices in y */
    int     n   /* number of nonzeros in y */
)
{
    int n5 ;
    register int j ;
    register short t ;
    t = 0 ;
    n5 = n % 5 ;
    for (j = 0; j < n5; j++) t += x [yi [j]] ;
    for (; j < n; j += 5)
    {
        t += x [yi[j]]
          +  x [yi[j+1]]
          +  x [yi[j+2]]
          +  x [yi[j+3]]
          +  x [yi[j+4]] ;
    }
    return (t) ;
}

/* sparse dot product x'y */
double SVMsparse_dot
(
    float  *x,  /* nonzero numerical values */
    short *xi,  /* nonzero indices in x */
    int   nzx,  /* number of nonzeros in x */
    float  *y,  /* nonzero numerical values */
    short *yi,  /* nonzero indices in y */
    int   nzy   /* number of nonzeros in y */
)
{
    register int i, j ;
    register double t ;
    t = 0. ;
    i = 0 ;
    j = 0 ;
    while ( (i < nzx) && (j < nzy) )
    {
        if ( yi [j] == xi [i] )
        {
            t += y [j]*x [i] ;
            i++ ;
            j++ ;
        }
        else
        {
            if ( yi [j] > xi [i] ) i++ ;
            else                   j++ ;
        }
    }
    return (t) ;
}

/* binary sparse dot product x'y (entries in x and y are 0 or 1) */
short SVMbinary_sparse_dot
(
    short *xi,  /* nonzero indices in x */
    int   nzx,  /* number of nonzeros in x */
    short *yi,  /* nonzero indices in y */
    int   nzy   /* number of nonzeros in y */
)
{
    register int i, j ;
    register short t ;
    t = 0 ;
    i = 0 ;
    j = 0 ;
    while ( (i < nzx) && (j < nzy) )
    {
        if ( yi [j] == xi [i] )
        {
            t++ ;
            i++ ;
            j++ ;
        }
        else
        {
            if ( yi [j] > xi [i] ) i++ ;
            else                   j++ ;
        }
    }
    return (t) ;
}

/* dense dot product x'x */
double SVMfloatxx
(
    float  *x,  /* nonzero numerical values */
    int     n   /* number of nonzeros in x */
)
{
    int n5 ;
    register int j ;
    register double t ;
    t = 0. ;
    n5 = n % 5 ;
    for (j = 0; j < n5; j++) t += (double) x [j]* (double) x [j] ;
    for (; j < n; j += 5)
    {
        t += x [j]*  x [j]
          +  x [j+1]*x [j+1]
          +  x [j+2]*x [j+2]
          +  x [j+3]*x [j+3]
          +  x [j+4]*x [j+4] ;
    }
    return (t) ;
}

/* =========================================================================
   === float_double_dot ======================================================
   =========================================================================
   Compute dot product of x and y
   ========================================================================= */
PRIVATE double SVMfloat_double_dot
(
    float  *x, /* first vector */
    double *y, /* second vector */
    int     n  /* length of vectors */
)
{
    int n5 ;
    register int j ;
    register double t ;
    t = 0. ;
    n5 = n % 5 ;
    for (j = 0; j < n5; j++) t += x [j]*y [j] ;
    for (; j < n; j += 5)
    {
        t += x [j]*y[j] + x [j+1]*y [j+1] + x [j+2]*y [j+2]
                        + x [j+3]*y [j+3] + x [j+4]*y [j+4] ;
    }
    return (t) ;
}

/* =========================================================================
   === short_gather ========================================================
   =========================================================================
   compress x by extracting the subspace elements and copying them to
   the compressed array y
   ========================================================================= */
PRIVATE void SVMshort_gather
(
    short *y, /* compressed vector */
    short *x, /* current x */
    int   *i, /* indices associated with subspace */
    int    n  /* number of elements to extract from x */
)
{
    int k, n5, *ik ;
    short *yk ;
    n5 = n % 5 ;
    for (k = 0; k < n5; k++) y [k] = x [i [k]] ;
    yk = y+k ;
    ik = i+k ;
    for (; k < n; k += 5)
    {
        *(yk++) = x [*(ik++)] ;
        *(yk++) = x [*(ik++)] ;
        *(yk++) = x [*(ik++)] ;
        *(yk++) = x [*(ik++)] ;
        *(yk++) = x [*(ik++)] ;
    }
}

/* =========================================================================
   === double_gather ========================================================
   =========================================================================
   compress x by extracting the subspace elements and copying them to
   the compressed array y
   ========================================================================= */
PRIVATE void SVMdouble_gather
(
    double *y, /* compressed vector */
    double *x, /* current x */
    int    *i, /* indices associated with subspace */
    int     n  /* number of elements to extract from x */
)
{
    int k, n5, *ik ;
    double *yk ;
    n5 = n % 5 ;
    for (k = 0; k < n5; k++) y [k] = x [i [k]] ;
    yk = y+k ;
    ik = i+k ;
    for (; k < n; k += 5)
    {
        *(yk++) = x [*(ik++)] ;
        *(yk++) = x [*(ik++)] ;
        *(yk++) = x [*(ik++)] ;
        *(yk++) = x [*(ik++)] ;
        *(yk++) = x [*(ik++)] ;
    }
}

/* =========================================================================
   === float_gather ========================================================
   =========================================================================
   compress x by extracting the subspace elements and copying them to
   the compressed array y
   ========================================================================= */
PRIVATE void SVMfloat_gather
(
    float *y, /* compressed vector */
    float *x, /* current x */
    int   *i, /* indices associated with subspace */
    int    n  /* number of elements to extract from x */
)
{
    int k, n5, *ik ;
    float *yk ;
    n5 = n % 5 ;
    for (k = 0; k < n5; k++) y [k] = x [i [k]] ;
    yk = y+k ;
    ik = i+k ;
    for (; k < n; k += 5)
    {
        *(yk++) = x [*(ik++)] ;
        *(yk++) = x [*(ik++)] ;
        *(yk++) = x [*(ik++)] ;
        *(yk++) = x [*(ik++)] ;
        *(yk++) = x [*(ik++)] ;
    }
}

/* =========================================================================
   === double_scatter ======================================================
   =========================================================================
   scatter elements from the compressed array y into x
   ========================================================================= */
PRIVATE void SVMdouble_scatter
(
    double *x, /* scatter y into x */
    double *y, /* compressed vector */
    int    *i, /* indices associated with subspace */
    int     n  /* number of elements to extract from x */
)
{
    int n5, *ik ;
    double *yk ;
    register int k ;
    n5 = n % 5 ;
    for (k = 0; k < n5; k++) x [i [k]] = y [k] ;
    yk = y+k ;
    ik = i+k ;
    for (; k < n; k += 5)
    {
        x [*(ik++)] = *(yk++) ;
        x [*(ik++)] = *(yk++) ;
        x [*(ik++)] = *(yk++) ;
        x [*(ik++)] = *(yk++) ;
        x [*(ik++)] = *(yk++) ;
    }
}

/* =========================================================================
   === float_scatter =======================================================
   =========================================================================
   scatter elements from the compressed array y into x
   ========================================================================= */
PRIVATE void SVMfloat_scatter
(
    float *x, /* scatter y into x */
    float *y, /* compressed vector */
    int   *i, /* indices associated with subspace */
    int    n  /* number of elements to extract from x */
)
{
    int n5, *ik ;
    float *yk ;
    register int k ;
    n5 = n % 5 ;
    for (k = 0; k < n5; k++) x [i [k]] = y [k] ;
    yk = y+k ;
    ik = i+k ;
    for (; k < n; k += 5)
    {
        x [*(ik++)] = *(yk++) ;
        x [*(ik++)] = *(yk++) ;
        x [*(ik++)] = *(yk++) ;
        x [*(ik++)] = *(yk++) ;
        x [*(ik++)] = *(yk++) ;
    }
}

/* =========================================================================
   === doubleCopy ==========================================================
   =========================================================================
   Copy the second vector to the first x = y
   ========================================================================= */
PRIVATE void SVMdoubleCopy
(
    double *x, /* copy of y */
    double *y, /* given vector */
    int     n  /* length of vectors */
)
{
    int n5 ;
    register int j ;
    n5 = n % 5 ;
    for (j = 0; j < n5; j++) x [j] = y [j] ;
    for (; j < n; j += 5)
    {
        x [j]   =  y[j] ;
        x [j+1] =  y[j+1] ;
        x [j+2] =  y[j+2] ;
        x [j+3] =  y[j+3] ;
        x [j+4] =  y[j+4] ;
    }
}

/* =========================================================================
   === floatCopy ===========================================================
   =========================================================================
   Copy the second vector to the first x = y
   ========================================================================= */
PRIVATE void SVMfloatCopy
(
    float *x, /* copy of y */
    float *y, /* given vector */
    int    n  /* length of vectors */
)
{
    int n5 ;
    register int j ;
    n5 = n % 5 ;
    for (j = 0; j < n5; j++) x [j] = y [j] ;
    for (; j < n; j += 5)
    {
        x [j]   =  y[j] ;
        x [j+1] =  y[j+1] ;
        x [j+2] =  y[j+2] ;
        x [j+3] =  y[j+3] ;
        x [j+4] =  y[j+4] ;
    }
}

/* =========================================================================
   === intCopy =============================================================
   =========================================================================
   Copy the second vector to the first x = y
   ========================================================================= */
PRIVATE void SVMintCopy
(
    int *x, /* copy of y */
    int *y, /* given vector */
    int  n  /* length of vectors */
)
{
    int n5 ;
    register int j ;
    n5 = n % 5 ;
    for (j = 0; j < n5; j++) x [j] = y [j] ;
    for (; j < n; j += 5)
    {
        x [j]   =  y[j] ;
        x [j+1] =  y[j+1] ;
        x [j+2] =  y[j+2] ;
        x [j+3] =  y[j+3] ;
        x [j+4] =  y[j+4] ;
    }
}

/* =========================================================================
   === shortCopy ===========================================================
   =========================================================================
   Copy the second vector to the first x = y
   ========================================================================= */
PRIVATE void SVMshortCopy
(
    short *x, /* copy of y */
    short *y, /* given vector */
    int    n  /* length of vectors */
)
{
    int n5 ;
    register int j ;
    n5 = n % 5 ;
    for (j = 0; j < n5; j++) x [j] = y [j] ;
    for (; j < n; j += 5)
    {
        x [j]   =  y[j] ;
        x [j+1] =  y[j+1] ;
        x [j+2] =  y[j+2] ;
        x [j+3] =  y[j+3] ;
        x [j+4] =  y[j+4] ;
    }
}

/* =========================================================================
   === indexCopy ===========================================================
   =========================================================================
   move elements in the array
   ========================================================================= */
PRIVATE void SVMindexCopy
(
    float    *x, /* target of copy is x */
    float    *y, /* copy from y */
    int   *iget, /* location in y to get element */
    int   *iput, /* location in x to put element */
    int       n  /* length of vectors */
)
{
    int n5 ;
    register int j ;
    n5 = n % 5 ;
    for (j = 0; j < n5; j++) x [iput [j]] = y [iget [j]] ;
    for (; j < n; j += 5)
    {
        x [iput [j]]   =  y[iget [j]] ;
        x [iput [j+1]] =  y[iget [j+1]] ;
        x [iput [j+2]] =  y[iget [j+2]] ;
        x [iput [j+3]] =  y[iget [j+3]] ;
        x [iput [j+4]] =  y[iget [j+4]] ;
    }
}

/* =========================================================================
   === indexCopyFlag =======================================================
   =========================================================================
   move elements in the array from x to y through z
   ========================================================================= */
PRIVATE void SVMindexCopyFlag
(
    float    *x, /* target of copy is x */
    float    *y, /* copy from y */
    int   *iget, /* location in y to get element */
    int   *iput, /* location in x to put element */
    int       n, /* length of vectors */
    short  flag  /* = 0 for collision free copy, +-1 when collisions possible */
)
{
    int n5 ;
    register int j, k ;
    if ( n == 0 ) return ;
    if ( flag == 0 ) SVMindexCopy (x, y, iget, iput, n) ;
    n5 = n % 5 ;
    if ( flag < 0 )
    {
        /* gather into x */
        for (j = 0; j < n5; j++) x [j] = y [iget [j]] ;
        for (; j < n; j += 5)
        {
            x [j]   =  y[iget [j]] ;
            x [j+1] =  y[iget [j+1]] ;
            x [j+2] =  y[iget [j+2]] ;
            x [j+3] =  y[iget [j+3]] ;
            x [j+4] =  y[iget [j+4]] ;
        }
        /* back scatter from x */
        k = n - n5 ;
        for (j = n-1; j >= k; j--) x [iput [j]] = x [j] ;
        for (; j > 3; j -= 5)
        {
            x [iput [j]]   =  x[j] ;
            x [iput [j-1]] =  x[j-1] ;
            x [iput [j-2]] =  x[j-2] ;
            x [iput [j-3]] =  x[j-3] ;
            x [iput [j-4]] =  x[j-4] ;
        }
    }
    else
    {
        /* gather into y */
        for (j = 0; j < n5; j++) y [j] = y [iget [j]] ;
        for (; j < n; j += 5)
        {
            y [j]   =  y[iget [j]] ;
            y [j+1] =  y[iget [j+1]] ;
            y [j+2] =  y[iget [j+2]] ;
            y [j+3] =  y[iget [j+3]] ;
            y [j+4] =  y[iget [j+4]] ;
        }
        /* back scatter from y into x */
        k = n - n5 ;
        for (j = n-1; j >= k; j--) x [iput [j]] = y [j] ;
        for (; j > 3; j -= 5)
        {
            x [iput [j]]   =  y[j] ;
            x [iput [j-1]] =  y[j-1] ;
            x [iput [j-2]] =  y[j-2] ;
            x [iput [j-3]] =  y[j-3] ;
            x [iput [j-4]] =  y[j-4] ;
        }
    }
}

/* =========================================================================
   === dense_diff ==========================================================
   =========================================================================
       set x = y - z
   ========================================================================= */
PRIVATE void SVMdense_diff
(
    double    *x,
    double    *y,
    double    *z,
    int        n  /* dimension */
)
{
    int n5 ;
    register int j ;
    n5 = n % 5 ;
    for (j = 0; j < n5; j++) x [j] = y [j] - z [j] ;
    for (; j < n; j += 5)
    {
        x [j  ] = y [j  ] - z [j]  ;
        x [j+1] = y [j+1] - z [j+1] ;
        x [j+2] = y [j+2] - z [j+2] ;
        x [j+3] = y [j+3] - z [j+3] ;
        x [j+4] = y [j+4] - z [j+4] ;
    }
}

/* =========================================================================
   === sparse_add ==========================================================
   =========================================================================
       set x = x + s*y, y is sparse
   ========================================================================= */
PRIVATE void SVMsparse_add
(
    double    *x,
    short    *yi,
    double     s, /* scalar */
    int        n  /* dimension */
)
{
    int n5 ;
    register int j ;
    n5 = n % 5 ;
    for (j = 0; j < n5; j++) x [yi [j]] += s ;
    for (; j < n; j += 5)
    {
        x [yi [j  ]] += s ;
        x [yi [j+1]] += s ;
        x [yi [j+2]] += s ;
        x [yi [j+3]] += s ;
        x [yi [j+4]] += s ;
    }
}

/* =========================================================================
   === sparse_daxpy ========================================================
   =========================================================================
       set x = x + s*y, y is sparse
   ========================================================================= */
PRIVATE void SVMsparse_daxpy
(
    double    *x,
    float     *y,
    short    *yi,
    double     s, /* scalar */
    int        n  /* dimension */
)
{
    int n5 ;
    register int j ;
    n5 = n % 5 ;
    for (j = 0; j < n5; j++) x [yi [j]] += s*y [j] ;
    for (; j < n; j += 5)
    {
        x [yi [j  ]] += s*y [j  ] ;
        x [yi [j+1]] += s*y [j+1] ;
        x [yi [j+2]] += s*y [j+2] ;
        x [yi [j+3]] += s*y [j+3] ;
        x [yi [j+4]] += s*y [j+4] ;
    }
}

/* =========================================================================
   === dense_daxpy =========================================================
   =========================================================================
       set x = x + s*y
   ========================================================================= */
PRIVATE void SVMdense_daxpy
(
    double    *x,
    float     *y,
    double     s, /* scalar */
    int        n  /* dimension */
)
{
    int n5 ;
    register int j ;
    n5 = n % 5 ;
    for (j = 0; j < n5; j++) x [j] += s*y [j] ;
    for (; j < n; j += 5)
    {
        x [j  ] += s*y [j  ] ;
        x [j+1] += s*y [j+1] ;
        x [j+2] += s*y [j+2] ;
        x [j+3] += s*y [j+3] ;
        x [j+4] += s*y [j+4] ;
    }
}

/* =========================================================================
   === int_init ============================================================
   =========================================================================
   Initialize an integer array
   ========================================================================= */
PRIVATE void SVMint_init
(
    short *x,  /* array to be initialized */
    short  s,  /* scalar */
    int    n   /* length of x */
)
{
    int j, n5 ;
    short *xj ;
    n5 = n % 5 ;
    for (j = 0; j < n5; j++) x [j] = s ;
    xj = x+j ;
    for (; j < n; j += 5)
    {
        *(xj++) = s ;
        *(xj++) = s ;
        *(xj++) = s ;
        *(xj++) = s ;
        *(xj++) = s ;
    }
}

PRIVATE void SVMscalar_init
(
    double  *x,  /* numerical values */
    double   s,  /* scalar */
    int      n   /* length of vectors */
)
{
    int j, n5 ;
    double *xj ;
    n5 = n % 5 ;
    for (j = 0; j < n5; j++) x [j] = s ;
    xj = x+j ;
    for (; j < n; j += 5)
    {
        *(xj++) = s ;
        *(xj++) = s ;
        *(xj++) = s ;
        *(xj++) = s ;
        *(xj++) = s ;
    }
}

/* =========================================================================
   === SVMadd ==============================================================
   =========================================================================
   add the components of a vector
   ========================================================================= */
PRIVATE double SVMadd
(
    double *x, /* vector */
    int     n  /* length of vector */
)
{
    int j, n5 ;
    double t ;

    t = SVMZERO ;
    n5 = n % 5 ;
    for (j = 0; j < n5; j++) t += x [j] ;
    for (; j < n; j += 5)
    {
        t += x [j] + x [j+1] + x [j+2] + x [j+3] + x [j+4] ;
    }
    return (t) ;
}

/* =========================================================================
   === SVMstep =============================================================
   =========================================================================
   set xnew = x + st*d
   ========================================================================= */
PRIVATE void SVMstep
(
    double *xnew, /* updated x vector */
    double    *x, /* current x */
    double    *d, /* search direction */
    double    st, /* stepsize */
    int        n  /* dimension */
)
{
    int j, n5 ;
    n5 = n % 5 ;
    if ( st == SVMONE )
    {
        for (j = 0; j < n5; j++) xnew [j] = x [j] + d [j] ;
        for (; j < n; j += 5)
        {
            xnew [j  ] = x [j  ] + d [j  ] ;
            xnew [j+1] = x [j+1] + d [j+1] ;
            xnew [j+2] = x [j+2] + d [j+2] ;
            xnew [j+3] = x [j+3] + d [j+3] ;
            xnew [j+4] = x [j+4] + d [j+4] ;
        }
    }
    else
    {
        for (j = 0; j < n5; j++)
        {
            xnew [j] = x [j] + st*d [j] ;
        }
        for (; j < n; j += 5)
        {
            xnew [j  ] = x [j  ] + st*d [j  ] ;
            xnew [j+1] = x [j+1] + st*d [j+1] ;
            xnew [j+2] = x [j+2] + st*d [j+2] ;
            xnew [j+3] = x [j+3] + st*d [j+3] ;
            xnew [j+4] = x [j+4] + st*d [j+4] ;
        }
    }
}

/* =========================================================================
   === SVMmax ==============================================================
   =========================================================================
   Return max {fabs (x [j]) : 1 <= j < n}
   ========================================================================= */
PRIVATE double SVMmax
(
    double *x,
    int     n
)
{
    register double xnorm ;
    register int j ;
    int n5 ;
    n5 = n % 5 ;
    xnorm = 0. ;
    for (j = 0; j < n5; j++) if ( xnorm < fabs (x [j]) ) xnorm = fabs (x [j]) ;
    for (; j < n; j += 5)
    {
        if ( xnorm < fabs (x [j]  ) ) xnorm = fabs (x [j]) ;
        if ( xnorm < fabs (x [j+1]) ) xnorm = fabs (x [j+1]) ;
        if ( xnorm < fabs (x [j+2]) ) xnorm = fabs (x [j+2]) ;
        if ( xnorm < fabs (x [j+3]) ) xnorm = fabs (x [j+3]) ;
        if ( xnorm < fabs (x [j+4]) ) xnorm = fabs (x [j+4]) ;
    }
    return (xnorm) ;
}

/* ==========================================================================
   === SVMpartialMinSort ====================================================
   ==========================================================================
   find indices of m smallest numbers in a list and arrange in a max heap
   ========================================================================== */
PRIVATE void SVMpartialMinSort
(
    int     *i, /* indices of m smallest numbers */
    double  *x, /* numbers to sort */
    int      m, /* length of i */
    int      n  /* length of x */
)
{
    int j ;
    double top ;
    if ( (m <= 0) || (n <= 0) ) return ;
    if ( m >= n )
    {
        for (j = 0; j < n; j++) i [j] = j ;
        SVMmaxheap_build (i-1, x, n) ;
        return ;
    }
    for (j = 0; j < m; j++) i [j] = j ;
    SVMmaxheap_build (i-1, x, m) ;
    top = x [i [0]] ;
    for (; j < n; j++)
    {
        if ( top > x [j] )
        {
            i [0] = j ;
            SVMmaxheapify (i-1, x, 1, m) ;
            top = x [i [0]] ;
        }
    }
}

/* ==========================================================================
   === SVMpartialMaxSort ====================================================
   ==========================================================================
   find indices of m largest numbers in a list
   ========================================================================== */
PRIVATE void SVMpartialMaxSort
(
    int    *i, /* indices of m largest numbers */
    double *x, /* numbers to sort */
    int     m, /* length of i */
    int     n  /* length of x */
)
{
    int j ;
    double bot ;
    if ( m >= n )
    {
        for (j = 0; j < n; j++) i [j] = j ;
        return ;
    }
    if ( (m <= 0) || (n <= 0) ) return ;
    for (j = 0; j < m; j++) i [j] = j ;
    SVMminheap_build (i-1, x, m) ;
    bot = x [i [0]] ;
    for (; j < n; j++)
    {
        if ( bot < x [j] )
        {
            i [0] = j ;
            SVMminheapify (i-1, x, 1, m) ;
            bot = x [i [0]] ;
        }
    }
}

/* ==========================================================================
   === SVMminHeapSort =======================================================
   ==========================================================================
   given a max heap, generate the list of indices corresponding to increasing
   order
   ========================================================================== */
PRIVATE void SVMminHeapSort
(
    int     *i, /* indices with increasing order based on x */
    int  *heap, /* min heap */
    double  *x, /* numbers to sort */
    int      n  /* size of heap */
)
{
    int j, k, l ;
    l = n ;
    for (j = 0; j < l; j++)
    {
        k = heap [0] ;
        n = SVMmaxheap_delete (heap-1, x, n) ;
        i [n] = k ;
    }
}

/* ==========================================================================
   === minheap_build ========================================================
   ==========================================================================
   build a min heap in heap [1..nheap]
   heap [i] is an index of an element of array x
   ========================================================================== */

PRIVATE void SVMminheap_build
(
    int  *heap, /* on input, an unsorted set of indices */
    double  *x, /* numbers to sort */
    int      n  /* number of elements to build into the heap */
)
{
    int p ;

    for (p = n/2 ; p >= 1 ; p--)
    {
        SVMminheapify (heap, x, p, n) ;
    }
}

/* ==========================================================================
   === minheap_delete =======================================================
   ==========================================================================
   delete the top element in a min heap
   ========================================================================== */
PRIVATE int SVMminheap_delete  /* return new size of heap */
(
    int  *heap, /* containing indices into x, 1..n on input */
    double  *x, /* not modified */
    int      n  /* number of items in heap */
)
{
    if (n <= 1)
    {
        return (0) ;
    }

    /* move element from the end of the heap to the top */
    heap [1] = heap [n] ;
    n-- ;
    SVMminheapify (heap, x, 1, n) ;
    return (n) ;
}

/* ========================================================================== */
/* === minheap_add ========================================================== */
/* ========================================================================== */
/* ========================================================================== */

/* add a new leaf to a min heap */

PRIVATE int SVMminheap_add
(
    int   leaf, /* the new leaf */
    int  *heap, /* size n, containing indices into x */
    double  *x, /* not modified */
    int      n  /* number of elements in heap not counting new one */
)
{
    int l, new, old ;
    double xold, xnew ;

    n++ ;
    old = n ;
    heap [old] = leaf ;
    xold = x [leaf] ;
    while ( old > 1 )
    {
        new = old/2 ;
        l = heap [new] ;
        xnew = x [l] ;
        if ( xnew > xold )
        { /* swap new and old */
            heap [new] = leaf ;
            heap [old] = l ;
        }
        else return (n) ;
        old = new ;
    }
    return (n) ;
}

/* ========================================================================== */
/* === minheapify =========================================================== */
/* ========================================================================== */

/* heapify starting at node p.  On input, the heap at node p satisfies the */
/* heap property, except for heap [p] itself.  On output, the whole heap */
/* satisfies the heap property. */

PRIVATE void SVMminheapify
(
    int *heap, /* size n, containing indices into x */
    double *x, /* not modified */
    int     p, /* start at node p in the heap */
    int     n  /* heap [1 ... n] is in use */
)
{
    int left, right, e, hleft, hright ;
    double xe, xleft, xright ;

    e = heap [p] ;
    xe = x [e] ;

    while ( SVMTRUE )
    {
        left = p * 2 ;
        right = left + 1 ;

        if ( right <= n )
        {
            hleft  = heap [left] ;
            hright = heap [right] ;
            xleft  = x [hleft] ;
            xright = x [hright] ;
            if (xleft < xright)
            {
                if (xe > xleft)
                {
                    heap [p] = hleft ;
                    p = left ;
                }
                else
                {
                    heap [p] = e ;
                    return ;
                }
            }
            else
            {
                if (xe > xright)
                {
                    heap [p] = hright ;
                    p = right ;
                }
                else
                {
                    heap [p] = e ;
                    return ;
                }
            }
        }
        else
        {
            if ( left <= n )
            {
                hleft = heap [left] ;
                xleft = x [hleft] ;
                if (xe > xleft)
                {
                    heap [p] = hleft ;
                    p = left ;
                }
            }
            heap [p] = e ;
            return ;
        }
    }
}

/* ========================================================================== */
/* === maxheap_build ======================================================== */
/* ========================================================================== */

/* build a max heap in heap [1..n] */

PRIVATE void SVMmaxheap_build
(
    int  *heap, /* on input, an unsorted set of elements */
    double  *x,
    int      n  /* number of elements to build into the heap */
)
{
    int p ;

    for (p = n/2 ; p >= 1 ; p--)
    {
        SVMmaxheapify (heap, x, p, n) ;
    }
}

/* ========================================================================== */
/* === maxheap_delete ======================================================= */
/* ========================================================================== */

/* delete the top element in a max heap */

PRIVATE int SVMmaxheap_delete  /* return new size of heap */
(
    int  *heap, /* containing indices into x, 1..n on input */
    double  *x, /* not modified */
    int      n  /* number of items in heap */
)
{
    if (n <= 1)
    {
        return (0) ;
    }

    /* move element from the end of the heap to the top */
    heap [1] = heap [n] ;
    n-- ;
    SVMmaxheapify (heap, x, 1, n) ;
    return (n) ;
}

/* ========================================================================== */
/* === maxheap_add ========================================================== */
/* ========================================================================== */

/* add a new leaf to a max heap */

PRIVATE int SVMmaxheap_add
(
    int   *heap, /* size n, containing indices into x */
    double   *x, /* not modified */
    int    leaf, /* the new leaf */
    int       n  /* number of elements in heap not counting new one */
)
{
    int l, new, old ;
    double xold, xnew ;

    n++ ;
    old = n ;
    heap [old] = leaf ;
    xold = x [leaf] ;
    while ( old > 1 )
    {
        new = old/2 ;
        l = heap [new] ;
        xnew = x [l] ;
        if ( xnew < xold )
        { /* swap new and old */
            heap [new] = leaf ;
            heap [old] = l ;
        }
        else return (n) ;
        old = new ;
    }
    return (n) ;
}

/* ========================================================================== */
/* === maxheapify =========================================================== */
/* ========================================================================== */

/* heapify starting at node p.  On input, the heap at node p satisfies the */
/* heap property, except for heap [p] itself.  On output, the whole heap */
/* satisfies the heap property. */

PRIVATE void SVMmaxheapify
(
    int *heap, /* size n, containing indices into x */
    double *x, /* not modified */
    int     p, /* start at node p in the heap */
    int     n  /* heap [1 ... n] is in use */
)
{
    int left, right, e, hleft, hright ;
    double xe, xleft, xright ;

    e = heap [p] ;
    xe = x [e] ;

    while ( SVMTRUE )
    {
        left = p * 2 ;
        right = left + 1 ;

        if (right <= n)
        {
            hleft  = heap [left] ;
            hright = heap [right] ;
            xleft  = x [hleft] ;
            xright = x [hright] ;
            if (xleft > xright)
            {
                if (xe < xleft)
                {
                    heap [p] = hleft ;
                    p = left ;
                }
                else
                {
                    heap [p] = e ;
                    return ;
                }
            }
            else
            {
                if (xe < xright)
                {
                    heap [p] = hright ;
                    p = right ;
                }
                else
                {
                    heap [p] = e ;
                    return ;
                }
            }
        }
        else
        {
            if (left <= n)
            {
                hleft = heap [left] ;
                xleft = x [hleft] ;
                if (xe < xleft)
                {
                    heap [p] = hleft ;
                    p = left ;
                }
            }
            heap [p] = e ;
            return ;
        }
    }
}

/* ==========================================================================
   === SVMnapsack ===========================================================
   ==========================================================================
    Find x that minimizes ||x-y|| while satisfying lo <= x <= hi, 1'x = b.
    The approach is to solve the dual problem obtained by introducing
    a multiplier lambda for the constraint 1'x = b.  The dual function is

    L (lambda) = min { .5||x-y||^2 + lambda (1'x - b): lo <= x <= hi }

    The dual function is concave. It is continuously differentiable.
    If mu denotes the maximizer of the dual function, then the solution
    of the primal problem is

    x = proj (y - mu*1) ,

    where proj (z) is the projection of z onto the set { x : lo <= x <= hi }.
    Thus we have

       proj (z)_i = lo_i if z_i <= lo_i,
                    hi_i if z_i >= hi_i,
                    z_i otherwise  .

    Note that for any lambda, the minimizing x in the dual function is

       x (lambda) = proj (y - lambda*1).

    The slope of the dual function is

      L'(lambda) = a'x(lambda) - b

    The solution technique is to start with an initial guess lambda for
    mu and search for a zero of L'. We have the following cases:

  1. L'(lambda) >= 0: mu >= lambda. If L' = 0, then done. Otherwise,
                      increase lambda using napup until slope vanishes

  2. L'(lambda) <= 0: mu <= lambda. If L' = 0, then done. Otherwise,
                      decrease lambda using napdown until slope vanishes

    The free set is those i for which lo_i < x_i (lambda) < hi_i.
    The active set is those i for which x_i = lo_i or x_i = hi_i.
    If we have a guess for the optimal active and free indices,
    then we can obtain a good guess for the starting lambda by
    setting the slope of the dual function to zero and solving for lambda.
    The guess for the active and free set is specified using the array B
    which has values:

        +1 => x_i = hi_i
        -1 => x_i = lo_i
         0 => lo_i < x_i < hi_i

    If there is no guess, then B is NULL.  Note that B is an INPUT array,
    it is not computed within the routine.

    The total time taken by napsack is O (n + h log n), where n is the size of
    y, h is the number of times an element of x (lambda) moves off the boundary
    into the free set plus the number of times elements move from the
    free set to the opposite boundary.  A heap is used to hold the entries
    in the boundary and in the free set.  If the slope vanishes at either
    the starting lambda, then no heap is constructed, and the time is
    just O (n).
   ========================================================================== */

PRIVATE double SVMnapsack /* return the final lambda */
(
    double         *x, /* holds y on input, and the solution x on output */
    double     lambda, /* starting guess for lambda */
    double          b, /* right side of constraint 1'x = b */
    double         lo, /* lower bounds */
    double         hi, /* upper bounds */
    int             n, /* problem dimension */
    double         *w, /* work array of length n */
    int          *iw1, /* work array of length n */
    int          *iw2  /* work array of length n */
)
{

    int j ;
    double xj, atx, a2sum, slope ;

    /* ---------------------------------------------------------------------- */
    /* compute the initial slope */
    /* ---------------------------------------------------------------------- */

    atx = SVMZERO ;
    a2sum = SVMZERO ;
    for (j = 0; j < n; j++)
    {
        xj = x [j] - lambda ;
        if      ( xj >= hi )  atx += hi ;
        else if ( xj <= lo )  atx += lo ;
        else                { atx += x [j] ; a2sum++ ; }
    }
    slope = atx - lambda*a2sum - b ; /* slope of the dual function */

    if      ( slope > SVMZERO ) /* case 1, increase lambda */
    {
        lambda = SVMnapup (x, b, lo, hi, n,
                           lambda, atx, a2sum, w, iw1-1, iw2-1);
    }
    else if ( slope < SVMZERO ) /* case 2, decrease lambda */
    {
        lambda = SVMnapdown (x, b, lo, hi, n,
                           lambda, atx, a2sum, w, iw1-1, iw2-1) ;
    }
    /* if slope = 0, then done */

    /* ---------------------------------------------------------------------- */
    /* replace x by x (lambda) */
    /* ---------------------------------------------------------------------- */
    for (j = 0; j < n; j++)
    {
        xj = x [j] - lambda ;
        if      ( xj < lo ) xj = lo ;
        else if ( xj > hi ) xj = hi ;
        x [j] = xj ;
    }

#ifdef DEBUG
    {
        asum = SVMZERO ;
        for (j = 0; j < n; j++) asum += x [j] ;
        if ( fabs(asum-b) > 1.e-8 )
        {
            DPRINT (("final x infeasible, a'x = %e != b (%e)\n", asum, b)) ;
            ERROR ("stop") ;
        }
    }
#endif

    return (lambda) ;
}

/* ========================================================================== */
/* === ASLnapup ============================================================= */
/* ========================================================================== */

/* Find x that minimizes ||x-y|| while satisfying the constraints
   lo <= x <= hi, a'x = b.  The algorithm is described in the napsack comments.
   It is assumed that the starting guess lambda for the dual multiplier is <=
   the correct multiplier. Hence, lambda will be increased.  The slope of
   the dual function is a'x - b. We increase lambda until a'x reaches b.
   If a_i > 0, then as lambda increases, x_i decreases. If a_i < 0, then
   as lambda increases, x_i increases. */

PRIVATE double SVMnapup /* return lambda */
(
    double        *y, /* input is target and output is solution */
    double         b, /* right side of constraint 1'x = b */
    double        lo, /* lower bounds */
    double        hi, /* upper bounds */
    int            n, /* problem dimension */
    double    lambda, /* initial guess for multiplier */
    double       atx, /* a'x */
    double     a2sum, /* sum_j a_j^2 */
    double *breakpts, /* break points */
    int  *bound_heap, /* work array */
    int   *free_heap  /* work array */
)
{
    int j, k, maxsteps, n_bound, n_free ;
    double minbound, minfree, new_break, slope, t, xj ;

    minbound = SVMINF ;
    minfree  = SVMINF ;

    /* -------------------------------------------------------------- */
    /* construct the heaps */
    /* -------------------------------------------------------------- */

    n_bound = 0 ;
    n_free = 0 ;
    for (j = 0; j < n; j++)
    {
        xj = y [j] - lambda ;
        if ( xj >= hi )
        {
            n_bound++ ;
            bound_heap [n_bound] = j ;
            t = y [j] - hi ;
            if ( minbound > t ) minbound = t ;
            breakpts [j] = t ;
        }
        else if ( xj > lo )
        {
            n_free++ ;
            free_heap [n_free] = j ;
            t = y [j] - lo ;
            if ( minfree > t ) minfree = t ;
            breakpts [j] = t ;
        }
    }

    maxsteps = 2*n + 1;
    for (k = 1 ; k <= maxsteps ; k++)
    {
        new_break = SVMMIN (minfree, minbound) ;
        slope = atx - new_break*a2sum - b ;
        if ( (slope <= SVMZERO) || (new_break == SVMINF) ) /* done */
        {
            if ( a2sum != SVMZERO ) lambda = (atx-b)/a2sum ;
            return (lambda) ;
        }
        lambda = new_break ;

        if ( k == 1 )
        {
            SVMminheap_build (free_heap, breakpts, n_free) ;
            SVMminheap_build (bound_heap, breakpts, n_bound) ;
        }

        /* -------------------------------------------------------------- */
        /* update the heaps */
        /* -------------------------------------------------------------- */

        if ( n_free > 0 )
        {
            while (breakpts [j = free_heap [1]] <= lambda)
            {
                a2sum-- ;
                atx +=  lo - y [j] ;
                n_free = SVMminheap_delete (free_heap, breakpts, n_free) ;
                if ( n_free == 0 ) break ;
            }
        }

        if ( n_bound > 0 )
        {
            while (breakpts [j = bound_heap [1]] <= lambda)
            {
                n_bound  = SVMminheap_delete (bound_heap, breakpts,n_bound);
                a2sum++ ;
                atx += y [j] - hi ;
                breakpts [j] = y [j] - lo ;
                n_free = SVMminheap_add (j, free_heap, breakpts, n_free) ;
                if ( n_bound == 0 ) break ;
            }
        }

        /*------------------------------------------------------------------- */
        /* get the smallest entry in each heap */
        /*------------------------------------------------------------------- */

        if ( n_free > 0 )  minfree = breakpts [free_heap [1]] ;
        else               minfree = SVMINF ;
        if ( n_bound > 0 ) minbound = breakpts [bound_heap [1]] ;
        else               minbound = SVMINF ;
    }
    /* this should not happen */
    error_handler ("Error in SVMnapup") ;
    return(lambda) ;
}

/* ========================================================================== */
/* === SVMnapdown =========================================================== */
/* ========================================================================== */

/* Find x that minimizes ||x-y|| while satisfying the constraints
   lo <= x <= hi, a'x = b.  The algorithm is described in the napsack comments.
   It is assumed that the starting guess lambda for the dual multiplier is >=
   the correct multiplier. Hence, lambda will be decreased.  The slope of
   the dual function is a'x - b. We decrease lambda until a'x reaches b.
   If a_i > 0, then as lambda decreases, x_i increases. If a_i < 0, then
   as lambda decreases, x_i decreases. */

PRIVATE double SVMnapdown /* return lambda */
(
    double        *y, /* input is target and output is solution */
    double         b, /* right side of constraint 1'x = b */
    double        lo, /* lower bounds */
    double        hi, /* upper bounds */
    int            n, /* problem dimension */
    double    lambda, /* initial guess for multiplier */
    double       atx, /* a'x */
    double     a2sum, /* sum_j a_j^2 */
    double *breakpts, /* break points */
    int  *bound_heap, /* work array */
    int   *free_heap  /* work array */
)
{
    int j, k, maxsteps, n_bound, n_free ;
    double maxbound, maxfree, new_break, slope, t, xj ;

    maxbound = -SVMINF ;
    maxfree  = -SVMINF ;

    /* -------------------------------------------------------------- */
    /* construct the heaps */
    /* -------------------------------------------------------------- */

    n_bound = 0 ;
    n_free = 0 ;
    for (j = 0; j < n; j++)
    {
        xj = y [j] - lambda ;
        if ( xj <= lo )
        {
            n_bound++ ;
            bound_heap [n_bound] = j ;
            t = y [j] - lo ;
            if ( maxbound < t ) maxbound = t ;
            breakpts [j] = t ;
        }
        else if ( xj < hi )
        {
            n_free++ ;
            free_heap [n_free] = j ;
            t = y [j] - hi ;
            if ( maxfree < t ) maxfree = t ;
            breakpts [j] = t ;
        }
    }

    maxsteps = 2*n + 1;
    for (k = 1; k <= maxsteps; k++)
    {
        new_break = SVMMAX (maxfree, maxbound) ;
        slope = atx - new_break*a2sum - b ;
        if ( (slope >= SVMZERO) || (new_break == -SVMINF) ) /* done */
        {
            if ( a2sum != SVMZERO ) lambda = (atx-b)/a2sum ;
            return (lambda) ;
        }
        lambda = new_break ;

        if ( k == 1 )
        {
            SVMmaxheap_build (free_heap, breakpts, n_free) ;
            SVMmaxheap_build (bound_heap, breakpts, n_bound) ;
        }

        /* -------------------------------------------------------------- */
        /* update the heaps */
        /* -------------------------------------------------------------- */

        if ( n_free > 0 )
        {
            while (breakpts [j = free_heap [1]] >= lambda)
            {
                a2sum-- ;
                atx +=  hi - y [j] ;
                n_free = SVMmaxheap_delete (free_heap, breakpts, n_free) ;
                if ( n_free == 0 ) break ;
            }
        }

        if ( n_bound > 0 )
        {
            while (breakpts [j = bound_heap [1]] >= lambda)
            {
                n_bound  = SVMmaxheap_delete (bound_heap, breakpts,n_bound);
                a2sum++ ;
                atx += y [j] - lo ;
                breakpts [j] = y [j] - hi ;
                n_free = SVMmaxheap_add (free_heap, breakpts, j, n_free) ;
                if ( n_bound == 0 ) break ;
            }
        }

        /*------------------------------------------------------------------- */
        /* get the biggest entry in each heap */
        /*------------------------------------------------------------------- */

        if ( n_free > 0 )  maxfree = breakpts [free_heap [1]] ;
        else               maxfree = -SVMINF ;
        if ( n_bound > 0 ) maxbound = breakpts [bound_heap [1]] ;
        else               maxbound = -SVMINF ;
    }
    /* this should not happen */
    error_handler ("Error in SVMnapup") ;
    return (lambda) ;
}

/* ========================================================================= */
/* === lp ==== min c'x subject to 1'x = b, 0 <= x <= C ===================== */
/* ========================================================================= */

PRIVATE double SVMlp /* return multiplier for linear constraint */
(
    double  *x,
    double  *c,
    double   C,
    double   b,
    int      n,
    double  mu, /* guess for multiplier */
    int     *i, /* work array of size n */
    int *index, /* work array of size n */
    double  *y  /* work array of size n */
)
{
    int j, k, l, m, nn, np, *indexP ;
    double slope, t ;

    /* Find the negative and positive components of c_i + mu.
       Store the indices of smallest components of c in bottom of
       array index, store indices of largest components of c in top
       of array index */
    nn = 0 ;
    np = n ;
    for (j = 0; j < n; j++)
    {
        if ( c [j] + mu < SVMZERO )
        {
            x [j] = C ;       /* tentatively set x_j = C */
            index [nn] = j ;
            nn++ ; /* count number of Cs in x */
        }
        else
        {
            np-- ;
            x [j] = SVMZERO ; /* tentatively set x_j = 0 */
            index [np] = j ;
        }
    }

    /* dual function L(mu) = min { c'x + mu(1'x-b) : 0 <= x <= C }
       compute the slope of the dual function */
    slope = (C*(double) nn) - b ;

    /* If slope is positive, then mu is increased and more components
       of x go to 0. If slope is negative, then mu is decreased and
       more components of x go to C.  As mu crosses break points,
       the slope changes by C.  Count the number of break points that
       must be crossed to make the slope of the dual function vanish. */
    m = ceil (fabs(slope)/C) ;

    if ( slope > SVMZERO ) /* increase mu, use nn indices, Cs switch to  0 */
    {
        /* extract components of c for sorting */
        for (j = 0; j < nn; j++) y [j] = c [index [j]] ;

        /* extract the m indices in lower part of index array corresponding to
           the largest components of cost vector c, the values of the
           solution x are all changed from C to 0 except for the index
           associated with the smallest c_i. The corrsponding x_i is set
           to a value between 0 and C so that the linear constraint is
           satisfied */

        SVMpartialMaxSort (i, y, m, nn) ;
        t = SVMINF ;
        for (j = 0; j < m; j++)
        {
            k = index [i [j]] ;
            x [k] = 0 ;
            if ( c [k] < t )
            {
                t = c [k] ;
                l = k ;
            }
        }
        x [l] = b - (nn-m)*C ;
        /* c_l + mu = 0, return mu = -c_l */
        return (-c [l]) ;
    }
    else if ( slope < SVMZERO ) /* decrease mu, use np indices, 0s switch to C*/
    {
        /* extract remaining components of c for sorting */
        for (j = np; j < n; j++) y [j] = c [index [j]] ;

        /* extract the m indices in upper part of index array corresponding
           to the smallest components of cost vector c, the values of the
           solution x are all changed from 0 to C except for the index
           associated with the largest c_i. The corrsponding x_i is set
           to a value between 0 and C so that the linear constraint is
           satisfied */

        SVMpartialMinSort (i, y+np, m, n-np) ;
        t = -SVMINF ;
        indexP = index+np ;
        for (j = 0; j < m; j++)
        {
            k = indexP [i [j]] ;
            x [k] = C ;
            if ( c [k] > t )
            {
                t = c [k] ;
                l = k ;
            }
        }
        x [l] = b - (nn+m-1)*C ;
        /* c_l + mu = 0, return mu = -c_l */
        return (-c [l]) ;
    }
    else return (mu) ; /* slope = 0 implies the guess for mu was correct */
}

/* ==========================================================================
   === SVMminSort ======================================================
   ==========================================================================
   Sort an array x in increasing order.
   The indices giving decreasing order are stored in y
   while x is not changed
   ========================================================================== */
PRIVATE void SVMminSort
(
    int *x,  /* numbers to sort (input), sorted numbers (output) */
    int *y,  /* work array of length n */
    int *z,  /* work array of length n+1 */
    int  n   /* number of elements to sort */
)
{
    int i, i1, i2, i1stop, i2stop, xstart, ystart, ix, iy, l,
        *xm, *xp, *yp, t ;
    /* find increasing sequences in x and store start points in z */
    if ( n <= 1 ) return ;
    if ( n == 2 )
    {
        if ( x [1] < x [0] )
        {
            t = x [0] ;
            x [0] = x [1] ;
            x [1] = t ;
            return ;
        }
    }

    xm = x-1 ;
    z [0] = 0 ;
    xstart = 1 ;
    for (i = 1; i < n; i++)
    {
        if ( xm [i] > x [i] )
        {
            z [xstart] = i ;
            xstart++ ;
        }
    }
    z [xstart] = n ;

    while ( 1 )
    {
        /* shuffle x into y, save break points */
        if ( xstart == 1 ) return ;      /* done */
        ystart = 0 ;                     /* new start */
        iy = 0 ;
        for (l = 0; l < xstart; )
        {
            z [ystart] = iy ;
            ystart++ ;
            i1 = z [l] ;
            l++ ;
            if ( l == xstart )           /* copy rest of x into y */
            {
                for (; i1 < n; i1++) y [i1] = x [i1] ;
                break ;
            }
            i2 = z [l] ;
            i1stop = i2 ;
            l++ ;
            i2stop = z [l] ;
            while ( 1 )
            {
                if ( x [i1] <= x [i2] )  /* pop x [i1] */
                {
                    y [iy] = x [i1] ;
                    iy++ ;
                    i1++ ;
                    if ( i1 == i1stop )  /* dump x [i2:i2stop] to y [iy: ] */
                    {
                        yp = y+(iy-i2) ;
                        iy += i2stop - i2 ;
                        for (; i2 < i2stop; i2++) yp [i2] = x [i2] ;
                        break ;
                    }
                }
                else                     /* pop x [i2] */
                {
                    y [iy] = x [i2] ;
                    iy++ ;
                    i2++ ;
                    if ( i2 == i2stop )  /* dump x [i1:i1stop] to y [iy: ] */
                    {
                        yp = y+(iy-i1) ;
                        iy += i1stop - i1 ;
                        for (; i1 < i1stop; i1++) yp [i1] = x [i1] ;
                        break ;
                    }
                }
            }
        }
        z [ystart] = n ;
        if ( ystart == 1 )               /* done, sorted numbers in y */
        {
            for (i = 0; i < n; i++) x [i] = y [i] ;
            return ;
        }

        /* shuffle y into x, save break points */
        xstart = 0 ;                     /* new start */
        ix = 0 ;
        for (l = 0; l < ystart; )
        {
            z [xstart] = ix ;
            xstart++ ;
            i1 = z [l] ;
            l++ ;
            if ( l == ystart )           /* copy rest of y into x */
            {
                for (; i1 < n; i1++) x [i1] = y [i1] ;
                break ;
            }
            i2 = z [l] ;
            i1stop = i2 ;
            l++ ;
            i2stop = z [l] ;
            while ( 1 )
            {
                if ( y [i1] <= y [i2] )  /* pop y [i1] */
                {
                    x [ix] = y [i1] ;
                    ix++ ;
                    i1++ ;
                    if ( i1 == i1stop )  /* dump y [i2:i2stop] to x [ix: ] */
                    {
                        xp = x+(ix-i2) ;
                        ix += i2stop - i2 ;
                        for (; i2 < i2stop; i2++) xp [i2] = y [i2] ;
                        break ;
                    }
                }
                else                     /* pop y [i2] */
                {
                    x [ix] = y [i2] ;
                    ix++ ;
                    i2++ ;
                    if ( i2 == i2stop )    /* dump y [i1:i1stop] queue */
                    {
                        xp = x+(ix-i1) ;
                        ix += i1stop - i1 ;
                        for (; i1 < i1stop; i1++) xp [i1] = y [i1] ;
                        break ;
                    }
                }
            }
        }
        z [xstart] = n ;
    }
}

/* ==========================================================================
   === SVMdefault ===========================================================
   ==========================================================================
       Set default parameter values.
   ========================================================================== */
void SVMdefault
(
    SVMparm *Parm, /* pointer to parameter structure */
    int  nfeature  /* number of features in the data */
)
{
    Parm->nsub = 450 ;   /* subspace size */
    /* When deciding which indices to put in the subspace, we attach greater
       weight to the indices associated with the previous subspace whose
       columns were saved. This weight factor, denoted PreAmp, essentially
       measures the time it takes to construct a column from scratch versus
       the time it takes to use a previously constructed column. */
    Parm->PreAmp = 2.e4 ;
    /* MaxKept is an upper bound on the fraction of indices from the
       previous subspace that are retained in the current subspace */
    Parm->MaxKept = .8 ;
    /* subspace error tolerance = MAX (err1, err2) */
    Parm->serr1 = .25 ;  /* err1 = serr1 times SVM tolerance */
    Parm->serr2 = .10 ;  /* err2 = serr2 times KKT error */
    /* if ASL performs no iterations, then we increase the convergence
       tolerance for force at least one iteration. The new error tolerance
       is the force factor time the current error in the subproblem. */
    Parm->ForceFac = .75 ;
    Parm->l_pert = 1.e-2 ;/* use index if Lagrangian error >= l_pert * SVM tol*/
    Parm->d_pert = 1.e-12 ; /* only include d_i in subspace if >= d_pert */
    Parm->bnd_pert = 1.e-3 ;/* threshhold for free variables is C*bnd_cut */
    Parm->lpfactor = 1.e4 ; /* parameter in decision rule of QP versus LP */
    /* default gamma is 1/nfeature */
    Parm->gamma = 1/ (double) nfeature ;
    /* default is that the data entries are not identically 1, if Ax_is_1 is true,
       then the numerical entries in the data are ignored */
    Parm->Ax_is_1 = SVMFALSE ;
    /* Parm->kernel passed down as subroutine argument */
    Parm->degree = 3 ;
    /* When building S or when updating the gradient, we need to compute
       the dot product between columns from Ax. When SparseDot is true, then
       this is the usual sparse dot product when vector. False means
       than one of the vectors is scattered into a vector of zero and the
       product is computed between the dense and the sparse vector. When
       the number of features is not too big, this can be more efficient
       since one fixed vector is multiplied against many vectors. */
    Parm->SparseDot = SVMFALSE ;
    Parm->maxouter = SVMINT_INF ;
    /* PrintLevel determines the amount of output
       0 = no output
       1 = print iteration number, error, and cost
       2 = output of level 1 + subspace statistics */
    Parm->PrintLevel = 1 ;
    Parm->FormS = SVMFALSE ; /* do not form the S matrix for linear kernels */
    Parm->PrintParms = SVMTRUE ;
}

/* ==========================================================================
   === SVMprint_parms =======================================================
   ==========================================================================
   print values in SVMparm structure
   ========================================================================== */
PRIVATE void SVMprint_parms
(
    SVMparm     *Parm /* SSMparm structure to be printed */
)
{
    /* numerical parameters */
    printf ("\nNumerical parameters:\n") ;
    printf ("parameter in kernels .............................. gamma: %e\n",
             Parm->gamma) ;
    printf ("degree for poly kernels .......................... degree: %i\n",
             Parm->degree) ;
    printf ("subspace tolerance <= serr1 times SVM tolerance ... serr1: %e\n",
             Parm->serr1) ;
    printf ("subspace tolerance <= serr2 times KKT error ....... serr2: %e\n",
             Parm->serr2) ;
    printf ("factor < 1 forcing at least one ASL iteration .. ForceFac: %e\n",
             Parm->ForceFac) ;
    printf ("use index if Lagrangian error >= l_pert*SVM tol .. l_pert: %e\n",
             Parm->l_pert) ;
    printf ("ignore index i if |d_i| <= d_pert ................ d_pert: %e\n",
             Parm->d_pert) ;
    printf ("threshold for free variables C*bndpert ......... bnd_pert: %e\n",
             Parm->bnd_pert) ;
    printf ("size of subspace ................................... nsub: %i\n",
             Parm->nsub) ;
    printf ("amplification factor for value of prior indices .. PreAmp: %e\n",
             Parm->PreAmp) ;
    printf ("maximum fraction of prior indices in new space .. MaxKept: %e\n",
             Parm->MaxKept) ;
    printf ("maximum number of outer iterations ............. maxouter: %i\n",
             Parm->maxouter) ;
    printf ("print level (0 = none, 2 = maximum) .......... PrintLevel: %i\n",
             Parm->PrintLevel) ;
    printf ("use lp for subspace if max_i |g_i| >= %e*C*lambda_k\n",
             Parm->lpfactor) ;

    /* logical parameters */
    printf ("\nLogical parameters:\n") ;
    if ( Parm->SparseDot )
        printf ("    Use sparse dot products\n") ;
    else
        printf ("    Use dense dot products\n") ;
    if ( Parm->Ax_is_1 )
        printf ("    All elements in data matrix are 1\n") ;
    if ( Parm->FormS )
        printf ("    Form dense submatrix if kernel = 0\n\n") ;
    else
        printf ("    Do not for dense submatrix if kernel = 0\n\n") ;
    if ( Parm->PrintParms )
        printf ("    Print the parameter structure\n") ;
    else
        printf ("    Do not print parameter structure\n") ;
}

PRIVATE double SVMtimer ( void )
{
    struct rusage ru ;
    double user, tsys ;

    (void) getrusage (RUSAGE_SELF, &ru) ;

    user =
    ru.ru_utime.tv_sec                  /* user time (seconds) */
    + 1e-6 * ru.ru_utime.tv_usec ;      /* user time (microseconds) */

    tsys =
    ru.ru_stime.tv_sec                  /* tsys time (seconds) */
    + 1e-6 * ru.ru_stime.tv_usec ;      /* tsys time (microseconds) */

    return (user + tsys) ;
}

PRIVATE void error_handler
(
    char *message
)
{
    fprintf (stderr, "Error: %s\n", message) ;
    abort ( ) ;
}

/* ========================================================================== */
/* ======= print_arrays ===================================================== */
/* ========================================================================== */

void print_arrays
(
    short       *x,
    char   *x_name,
    int     length
)
{
    int i ;

    printf ("%s = [\n", x_name) ;
    for (i = 0; i < length; i++)
    {
        printf ("%i\n", x [i]) ;
    }
    printf ("] ;\n") ;
}

/* ========================================================================== */
/* ======= print_arrayi ===================================================== */
/* ========================================================================== */

void print_arrayi
(
    int         *x,
    char   *x_name,
    int     length
)
{
    int i ;

    printf ("%s = [\n", x_name) ;
    for (i = 0; i < length; i++)
    {
        printf ("%i\n", x [i]) ;
    }
    printf ("] ;\n") ;
}

/* ========================================================================== */
/* ======= print_arrayx ===================================================== */
/* ========================================================================== */

void print_arrayx
(
    float       *x,
    char   *x_name,
    int     length
)
{
    int i ;

    printf ("%s = [\n", x_name) ;
    for (i = 0; i < length; i++)
    {
        printf ("%30.20e\n", x [i]) ;
    }
    printf ("] ;\n") ;
}

/* ========================================================================== */
/* ======= print_arrayd ===================================================== */
/* ========================================================================== */

void print_arrayd
(
    double      *x,
    char   *x_name,
    int     length
)
{
    int i ;

    printf ("%s = [\n", x_name) ;
    for (i = 0; i < length; i++)
    {
        printf ("%30.20e\n", x [i]) ;
    }
    printf ("] ;\n") ;
}

void print_matrix
(
    short  *Ai,
    int    *Ap,
    float  *Ax,
    char *name,
    int      n
)
{
    int j, k ;

    printf ("%s = [\n", name) ;
    if ( Ax == NULL )
    {
        for (j = 0; j < n; j++)
        {
            for (k = Ap [j]; k < Ap [j+1]; k++)
            {
                printf ("%i %i %e\n", Ai [k], j, 1.) ;
            }
        }
    }
    else
    {
        for (j = 0; j < n; j++)
        {
            for (k = Ap [j]; k < Ap [j+1]; k++)
            {
                printf ("%i %i %e\n", Ai [k], j, Ax [k]) ;
            }
        }
    }
    printf ("] ;\n") ;
}

